// CPPConst.h: constant definitions used in the C++ parser

#ifndef INCL_CPP_CONST_H
#define INCL_CPP_CONST_H

// For platforms that have no bool, include something from STL
// to pull in the bool definition that they must have
#include <new>

#include "JLSmartFlag.h"

// I use a class here, since namespaces aren't widely available.

// Use "smart flags" for the flag sets, because they force
// initialization, and restrict operations on the flags.

class CPPConst {
private:
   // Random enumeration used to make template instantiations
   // of SmartFlag<> incompatible with each other
   enum Signatures { 
      sig1=1048, sig2=1984, sig3=8013, sig4=1287, sig5=1237, 
      sig6=1289, sig7=3712, sig8=8937, sig9=9812, sig10=7312
   };
public:
   typedef SmartFlag<0x10000, sig1> TypeSpecifier;
   static const TypeSpecifier tsNONE;       // 0x0L
   static const TypeSpecifier tsVOID;       // 0x1L
   static const TypeSpecifier tsCHAR;       // 0x2L
   static const TypeSpecifier tsSHORT;      // 0x4L
   static const TypeSpecifier tsINT;        // 0x8L
   static const TypeSpecifier tsLONG;       // 0x10L
   static const TypeSpecifier tsFLOAT;      // 0x20L
   static const TypeSpecifier tsDOUBLE;     // 0x40L
   static const TypeSpecifier tsSIGNED;     // 0x80L
   static const TypeSpecifier tsUNSIGNED;   // 0x100L
   static const TypeSpecifier tsTYPEID;     // 0x200L
   static const TypeSpecifier tsSTRUCT;     // 0x400L
   static const TypeSpecifier tsENUM;       // 0x800L
   static const TypeSpecifier tsUNION;      // 0x1000L
   static const TypeSpecifier tsCLASS;      // 0x2000L
   static const TypeSpecifier tsWCHAR_T;    // 0x4000L
   static const TypeSpecifier tsBOOL;       // 0x8000L
   static const TypeSpecifier tsINT64;      // 0x10000L

   typedef SmartFlag<0x2, sig2> CVQualifier;
   static const CVQualifier cvNONE;       // 0x0
   static const CVQualifier cvCONST;      // 0x1
   static const CVQualifier cvVOLATILE;   // 0x2

   typedef SmartFlag<0x80, sig3> StorageClass;
   static const StorageClass scNONE;        // 0x0
   static const StorageClass scAUTO;        // 0x1
   static const StorageClass scREGISTER;    // 0x2
   static const StorageClass scSTATIC;      // 0x4
   static const StorageClass scEXTERN;      // 0x8
   static const StorageClass scTYPEDEF;     // 0x10
   // Compiler-specific extensions
   static const StorageClass scMSDLLImport; // 0x20
   static const StorageClass scMSDLLExport; // 0x40
   static const StorageClass scMSBased;     // 0x80

   // Declaration Modifiers
   typedef SmartFlag<0x8, sig3> DeclModifier;
   static const DeclModifier dmNONE;          // 0x0
   static const DeclModifier dmVIRTUAL;       // 0x1
   static const DeclModifier dmINLINE;        // 0x2
   static const DeclModifier dmFRIEND;        // 0x4
   static const DeclModifier dmEXPLICIT;      // 0x8

   // Calling convention
   typedef SmartFlag<0x2, sig4> CallingConvention;
   // default
   static const CallingConvention ccCDECL;         // 0x0
   // Compiler-specific extensions
   static const CallingConvention ccMS_FASTCALL;   // 0x1
   static const CallingConvention ccMS_STDCALL;    // 0x2

   // Access specifiers
   typedef SmartFlag<0x4, sig5> AccessSpecifier;
   static const AccessSpecifier asNONE;      // 0x0
   static const AccessSpecifier asPUBLIC;    // 0x1
   static const AccessSpecifier asPROTECTED; // 0x2
   static const AccessSpecifier asPRIVATE;   // 0x4

   // Base specifiers
   typedef SmartFlag<0x8, sig6> BaseSpecifier;
   static const BaseSpecifier bsNONE;      // 0x0
   static const BaseSpecifier bsPUBLIC;    // 0x1
   static const BaseSpecifier bsPROTECTED; // 0x2
   static const BaseSpecifier bsPRIVATE;   // 0x4
   static const BaseSpecifier bsVIRTUAL;   // 0x8
};

#endif
