// CPPDecl.cpp: C++ Declaration base class

#include "CPPDecl.h"

////////////////////////////////////////////////////////////////////
// Constructor
////////////////////////////////////////////////////////////////////
CPPDecl::CPPDecl(
   const JLStr& name_,
   CPPConst::AccessSpecifier accessSpecifier_
) :
   name(name_),
   accessSpecifier(accessSpecifier_)
   #ifndef NDEBUG
      , magic(CPPDeclMagic)
   #endif
{}

////////////////////////////////////////////////////////////////////
// Destructor
////////////////////////////////////////////////////////////////////
CPPDecl::~CPPDecl()
{
   assertValid();
   #ifndef NDEBUG
      magic = 0;
   #endif
   // Do not delete "scope" -- it is the parent!
}


////////////////////////////////////////////////////////////////////
// SameDecl: Is the declaration identicial to another declaration?
// Inputs: none
// Outputs: none
// Return: 
//    bool    true if the declarations are identical.  CPPDecl
//            only checks the access specifiers and flavor.
//            Derived classes should call this and perform 
//            further checks.
//
// Note: The name of a declaration must be compared elsewhere, as
// that information is only stored in the declaration map of the 
// scope.
////////////////////////////////////////////////////////////////////
// virtual
bool 
CPPDecl::SameDecl(
   const CPPDecl& rhs
) const 
{
   assertValid(); 
   return 
      IsType() == rhs.IsType() &&
      GetAccessSpecifier() == rhs.GetAccessSpecifier();
}

