// CPPParserSym.h: Subclass of CPPParser that has symbol table support
// This class implements two things:
//   1) Overrides of parser base class pure-virtual methods 
//      for accessing the symbol table.
//   2) Operational functionality that is independent of the grammar.
//

#ifndef INCL_CPPPARSER_SYM_H
#define INCL_CPPPARSER_SYM_H

#include "CPPParser.h"
#include "CPPSymTableIncomplete.h"

class CPPParserSym : public CPPParser
{
public:
   DECLARE_INCOMPLETE(OptionSet);

   // ctor
	CPPParserSym(ANTLRTokenBuffer *input);
   // ctor for template specialization
	CPPParserSym(
      ANTLRTokenBuffer *input,
      CPPSymTableRefI symTable_,
      OptionSetRefI optionSet_
   );
   // init
   virtual void init();
   // dtor
   virtual ~CPPParserSym();

   // Debugging
   void DumpScopes() const;

   // Get/set parser options
   virtual void SetOption(Option option, bool value);
   virtual bool GetOption(Option option) const;
   OptionSetRefI GetOptions() const { return options; }

protected:
   
   //
   // Overrides of the parser base class
   // Most symbol-table management methods are
   // simply forwarded to the symbol table object.
   // Get definitions of pasrer virtual methods from a header file.
   // 
   #define PARSER_PURE_VIRTUAL
   #include "CPPParserVirtual.h"
   #undef PARSER_PURE_VIRTUAL

   // consume the next token
   virtual void consume();

   // Process a #pragma pack()...
   void CPPParserSym::HandleMSPacking();

private:
   //
   // Helper methods 
   //

   // set for holding option values
   OptionSetRefI options;
   // Internal method for creating an option set
   static OptionSetRefI NewOptionSet();

   // Pointer to symbol table
   CPPSymTableRefI symTable;

   // Stack of token lists that are collecting tokens as
   // they are consumed.  Every list on the stack will 
   // have an additional token added to the end whenever
   // a token is consumed by the parser.
   TokenListListRefI tokenCollectionLists;

   // Stack of integers that are used to store the input token
   // marks as they are made during guess mode, for the purpose of 
   // removing tokens from the token collection lists when a 
   // rewind is made.
   DECLARE_INCOMPLETE(MarkerStack);
   MarkerStackRefI markerStack;
   static MarkerStackRefI NewMarkerStack();
};

#endif
