// CPPScopeLevel.h: Scope stack entry definition
#ifndef INCL_CPPSCOPELEVEL_H
#define INCL_CPPSCOPELEVEL_H

#include "CPPScopeIncomplete.h"

///////////////////////////////////////////////////////////////////////////////
// Class that is entered on the scope stack when a new scope is entered.
// The "scope" member is the scope that is entered.
// The "resolved" member, if true, indicates that searching of the scope
// stack should stop here.  The "resolved" member is mostly used for
// qualified names, e.g.:
//     class A { class B { int f(); }; };
//
// when "class A {" is seen, class A's (new) scope is pushed on the scope
// stack with "resolved" set to false; similarly for "class B {".  However,
// in the case of:
//     int A::B::f() {...}
//
// when when the A:: is seen, class A's scope is pushed on the stack
// with "resolved" set to true, meaning that A's scope is the only scope
// to be searched; similarly for B::.
///////////////////////////////////////////////////////////////////////////////
class CPPScopeLevel
{
public:
   CPPScopeLevel(
      CPPScopeRefI scope_ = CPPScopeRefI(),
      bool resolved_ = false
   ) :
      scope(scope_),
      resolved(resolved_)
   {}
   CPPScopeRefI GetScope() const { return scope; }
   bool GetResolved() const { return resolved; }
private:
   CPPScopeRefI scope;
   bool resolved;
};

#endif
