// CPPScopeSaver.cpp: Implementation of the ScopeSaver class of the parser.

#include "CPPParser.h"
#include "CPPScopeContainers.h"

CPPParser::ScopeSaver::ScopeSaver(
   CPPParser* parser_,
   CPPScopeState state_
) : 
   parser(parser_),
   state(state_),
   committed(false)
{
   // Copy all of the scopes exceeeding the reference scope 
   // state into a separate stack.
   scopeStack = new CPPScopeStack;

   // This is not especially efficient because it pops the scopes and 
   // then puts them back...  could use improvement.
   while (parser->SaveScope() != state)
   {
      scopeStack->push_front(parser->PopScope());
   }
   Restore();
}

void 
CPPParser::ScopeSaver::Restore()
{
   // Pop the scopes exceeding the given state, 
   // and restore the original scope stack.
   parser->RestoreScope(state);
   for (
      CPPScopeStack::iterator iter = scopeStack->begin();
      iter != scopeStack->end();
      iter++
   )
   {
      parser->PushScope(*iter);
   }
}

