// CPPSymTable.cpp: init/destroy for C++ parser symbol table

#include "CPPSymTable.h"
#include "CPPScopeNormal.h"
#include "CPPType.h"
#include "CPPTypeContainers.h"
#include "CPPScopeContainers.h"
#include "TokenList.h"

CPPSymTable::CPPSymTable(
   CPPParserSym* parser_
) :
   parser(parser_),
   templateLevel(0)
{
   // Create data structures
   builtinTypes = new CPPBuiltinTypeTable;
   scopeStack = new CPPScopeStack;
   globalScope = new CPPScopeNormal(CPPScope::ScopeGlobal, "");
   emptyScope = new CPPScopeNormal(CPPScope::ScopeLocal, "");

   // Initialize packing stack to contain the default value
   packingStack = NewPackingStack();
   PushPacking(DefaultPacking);

   // Initialize the map from template parameter idx/template level to 
   // template placeholder types and values
   templatePlaceholderTypeMap = MakeTemplatePlaceholderTypeMap();
   templatePlaceholderValueMap = MakeTemplatePlaceholderValueMap();

   // Push the global scope on the scope stack
   scopeStack->push_front(CPPScopeLevel(globalScope));
}


CPPSymTable::~CPPSymTable()
{
}
