// CPPSymTableMisc.cpp: misc utility methods

#include "CPPSymTable.h"
#include "CPPTypeTemplated.h"
#include "CPPScope.h"

/////////////////////////////////////////////////////////////////////
// NameIsDecoratedName: See if the name is equivalent to the
// decorated name.  "Equivalent to" means that a name is
// considered the same as a decorated template name using
// the same prefix.
//
// Inputs:
//    const JLStr&   name           The name to compare
//    const JLStr&   decoratedName  The decorated name to compare
/////////////////////////////////////////////////////////////////////
//static 
bool 
CPPSymTable::NameIsDecoratedName(
   const JLStr& name, 
   const JLStr& decoratedName
)
{
   if (name == decoratedName)
   {
      return true;
   } else {
      JLStr::size_type tmpPos = decoratedName.find('<');
      if (tmpPos != JLStr::npos)
      {
         return name == JLStr(decoratedName.begin(), decoratedName.begin() + tmpPos);
      } else {
         return false;
      }
   }
}


/////////////////////////////////////////////////////////////////////
// GetUniqueName: Get a name that is unique to the current scope
//
// Inputs: none
// Outputs: none
// Return:
//    JLStr    A name that is guaranteed to be unique.
/////////////////////////////////////////////////////////////////////
JLStr 
CPPSymTable::GetUniqueName() const
{
   return CPPScope::UniqueName();
}


/////////////////////////////////////////////////////////////////////
// IsGeneratedName: test to see if a name is computer-generated
//
// Inputs:
//    const JLStr&    name       The name to test
// Outputs: none
// Return:
//    bool     true if the name was generated from GetUniqueName()
/////////////////////////////////////////////////////////////////////
bool
CPPSymTable::IsGeneratedName(
   const JLStr& name
) const
{
   return CPPScope::IsGeneratedName(name);
}


