// CPPSymTableTemplate.cpp: template processing for parser

#include "CPPSymTable.h"
#include "CPPTypeClass.h"
#include "CPPTypeTemplated.h"
#include "CPPTypeTemplatePlaceholder.h"
#include "CPPScope.h"
#include "CPPDecl.h"
#include "CPPExpr.h"
#include "TemplateArgumentList.h"
#include "TemplateParameterList.h"


///////////////////////////////////////////////////////////////////
// SetTokensInTemplateDeclaration:  Set the tokens for the body of
// a templated declaration, so that it can be reparsed when needed.
//
// Inputs:
//    CPPTypeRefI                 templateType   The templated type
//    TemplateParameterListRefI   templateParameters 
//                                               The template parameters
//    TokenListRefI               tokens         The token list
//   
///////////////////////////////////////////////////////////////////
void 
CPPSymTable::SetTokensInTemplateDeclaration(
   CPPTypeRefI templateType,
   TemplateParameterListRefI templateParameters,
   TokenListRefI tokens
)
{
   // Don't store the tokens for fully-specialized templates
   if (templateParameters->size() != 0)
   {
      assert(templateType->GetFlavor() == CPPType::Templated);
      STATIC_CAST(CPPTypeTemplated*, templateType.get())->SetTokens(tokens);
   }
}


///////////////////////////////////////////////////////////////////
// TypeUsesInnerTemplateParms: Determine if a type uses (or is)
// template placeholder type from the innermost template head.
//
// Inputs:
//    CPPTypeRefI    type     The type to check
// Outputs: none
// Return:
//    bool     true if the type uses (or is) a template placeholder
//             type from the innermost template head.
///////////////////////////////////////////////////////////////////
bool 
CPPSymTable::TypeUsesInnerTemplateParms(
   CPPTypeRefI type
)
{
   // Compare the template nesting level of the placeholder type with
   // one less than the current tempalte nesting level (because the
   // nesting level is bumped immediately following the template head).
   CPPTypeRefI placeholderType;
   return 
      type->FindTemplatePlaceholder(placeholderType) &&
      (
         STATIC_CAST(CPPTypeTemplatePlaceholder*, placeholderType.get())->GetTemplateLevel() == 
         GetTemplateLevel() - 1
      );
}
