// CPPSymTableTypeTemplate.cpp: Template type manipulation

#include "CPPSymTable.h"
#include "CPPScopeTemplate.h"
#include "CPPTypeTemplatePlaceholder.h"
#include "CPPDecl.h"
#include "TemplateParameterList.h"

/////////////////////////////////////////////////////////////////////
// DeclareTemplateTypeParameter: Declare a template "type" parameter,
// for example, "T" in:
//     template <class T> ...
// 
// Inputs:
//    const JLStr&           name                The name of the class.
//    const ANTLRTokenPtr&   errTok              A token for error-handling
//    TokenListRefI          defaultTokens       The tokens that comprise the
//                                               default type of the parameter
//    short                  parameterIdx        The index of the template
//                                               parameter in its list.
//    short                  templateLevel       The template nesting level.
// Outputs:
//    TemplateParameterRefI& templateParmReturn  The template type parameter
//                                               entry that was createed.
// Return: none
/////////////////////////////////////////////////////////////////////
void
CPPSymTable::DeclareTemplateTypeParameter(
   const JLStr& name,
   TokenListRefI defaultTokens,
   short parameterIdx,
   short templateLevel,
   TemplateParameterRefI& templateParmReturn
)
{
   JLStr tmpName = (name.length() == 0) ? GetUniqueName() : name;
   CPPTypeRefI placeholderType = GetTemplatePlaceholderType(parameterIdx, templateLevel);
   templateParmReturn = 
      new TemplateParameterType(tmpName, defaultTokens, placeholderType);

   // enter the template argument into the current scope
   CPPDeclRefI decl = new CPPDeclType(
      tmpName,
      CPPConst::asPUBLIC,
      placeholderType,
      false    // not a typedef
   );
   GetCurrentScope()->AddDecl(decl);
}
