// CPPSymTableValueTemplate.cpp: Template value manipulation

#include "CPPSymTable.h"
#include "CPPScope.h"
#include "CPPExpr.h"
#include "CPPDecl.h"
#include "TemplateParameterList.h"

/////////////////////////////////////////////////////////////////////
// DeclareTemplateValueParameter: Declare a template "value" parameter,
// for example, "I" in:
//     template <int I> ...
// 
// Inputs:
//    const JLStr&           name                The name of the class.
//    const ANTLRTokenPtr&   errTok              A token for error-handling
//    TokenListRefI          defaultTokens       The tokens that comprise the
//                                               default type of the parameter
//    short                  parameterIdx        The index of the template
//                                               parameter in its list.
//    short                  templateLevel       The template nesting level.
//    CPPTypeRefI            parameterType       The type of the value placeholder
// Outputs:
//    TemplateParameterRefI& templateParmReturn  The template value parameter
//                                               entry that was createed.
// Return: none
/////////////////////////////////////////////////////////////////////
void
CPPSymTable::DeclareTemplateValueParameter(
   const JLStr& name,
   TokenListRefI defaultTokens,
   short parameterIdx,
   short templateLevel,
   CPPTypeRefI parameterType,
   TemplateParameterRefI& templateParmReturn
)
{
   JLStr tmpName = (name.length() == 0) ? GetUniqueName() : name;

   // TODO: does the placeholder expression need the type?
   CPPExprRefI placeholderExpr = GetTemplatePlaceholderValue(parameterIdx, templateLevel);

   // Create the template value parameter
   templateParmReturn = new TemplateParameterValue(
      tmpName,
      defaultTokens,
      parameterType,
      placeholderExpr
   ); 

   // enter the template argument into the current scope
   CPPDeclRefI decl = new CPPDeclObject(
      tmpName,
      CPPConst::asPUBLIC,
      parameterType,
      CPPConst::scNONE,
      CPPConst::dmNONE,
      placeholderExpr,
      StrListRefI()
   );

   GetCurrentScope()->AddDecl(decl);
}
