// CPPType.cpp: C++ Declaration base class

#include "CPPType.h"
#include "CPPTypeContainers.h"

////////////////////////////////////////////////////////////////////
// Static data
////////////////////////////////////////////////////////////////////
CPPTypePtrSetCRefI CPPType::emptyTypeSet = new CPPTypePtrSet;

////////////////////////////////////////////////////////////////////
// Constructor
////////////////////////////////////////////////////////////////////
CPPType::CPPType(
   Flavor flavor_
) :
   flavor(flavor_)
   #ifndef NDEBUG
      , magic(CPPTypeMagic)
   #endif
{}

////////////////////////////////////////////////////////////////////
// Destructor
////////////////////////////////////////////////////////////////////
CPPType::~CPPType()
{
   assertValid();
   #ifndef NDEBUG
      magic = 0;
      flavor = Invalid;
   #endif
}

////////////////////////////////////////////////////////////////////
// AddDerivedType: Add a derived type to the list of derived types.
//
// Inputs:
//    CPPType*    derivedType    The derived type to add.
////////////////////////////////////////////////////////////////////
void 
CPPType::AddDerivedType(
   CPPType* derivedType
)
{
   assertValid();
   
   // Derived type must be a modifier
   assert(
      derivedType->flavor != Enum &&
      derivedType->flavor != Class &&
      derivedType->flavor != Builtin &&
      derivedType->flavor != TemplatePlaceholder
   );

   if (derivedTypes.get() == 0)
   {
      derivedTypes = new CPPTypePtrSet;
   }
   derivedTypes->insert(derivedType);
}

////////////////////////////////////////////////////////////////////
// RemoveDerivedType: Remove a derived type from the list of 
// derived types.  
//
// Inputs:
//    CPPType*    derivedType    The derived type to remove.
//
// NOTE: It is OK to attempt removal of a type modifier that is
// not in the derived type list -- it will be ignored.
// This sometimes happens when temporary type modifierss are created
// and destroyed for the purpose of comparison to existing modifiers.
////////////////////////////////////////////////////////////////////
void 
CPPType::RemoveDerivedType(
   CPPType* derivedType
)
{
   // Derived type must be a modifier
   assert(
      derivedType->flavor != Enum &&
      derivedType->flavor != Class &&
      derivedType->flavor != Builtin &&
      derivedType->flavor != TemplatePlaceholder
   );

   if (derivedTypes.get() != 0)
   {
      CPPTypePtrSet::iterator iter = derivedTypes->find(derivedType);
      if (iter != derivedTypes->end())
      {
         derivedTypes->erase(iter);
      }
   }
}

////////////////////////////////////////////////////////////////////
// GetDerivedTypes: Get the list of derived types.  Note that
// if the list of derived types is empty, then this will return
// the address of a static empty list, so the caller should not
// hang on to the resulting pointer.
////////////////////////////////////////////////////////////////////
CPPTypePtrSetCRefI
CPPType::GetDerivedTypes() const
{
   if (derivedTypes.get() == 0)
   {
      return emptyTypeSet;
   } else {
      return derivedTypes;
   }
}
