// CPPTypeArray.cpp: Implementation for array type modifier.

#include "CPPType.h"
#include "CPPTypeArray.h"
#include "stdio.h"

/////////////////////////////////////////////////////////////////////////
// ctor/dtor
/////////////////////////////////////////////////////////////////////////
CPPTypeArray::CPPTypeArray(
   CPPTypeRefI baseType_,
   int dimension_,
   bool addToBase
) :
   CPPTypeModifier(Array, baseType_),
   dimension(dimension_)
{
   assert(baseType_.get() != 0);
   if (addToBase)
   {
      AddToBase();
   }
}

//virtual
CPPTypeArray::~CPPTypeArray()
{ 
   assertValid(); 
   RemoveFromBase(); 
}

/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                             to format the declarator name
//                             in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeArray::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI,
   bool forSpecialization
) const
{
   assertValid();
   
   JLStr modifiedSuffix;
   if (dimension != 0)
   {
      char buf[20];
      sprintf(buf, "[%d]", dimension);
      modifiedSuffix = buf;
   } else {
      modifiedSuffix = "[]";
   }
   modifiedSuffix += suffixStr;

   return GetBaseType()->FormatName(
      declName, 
      modifiedSuffix, 
      StrListRefI(), 
      forSpecialization
   );
}
