// CPPTypeArray.h: definition of array-of type modifier class

#ifndef INCL_CPPTYPEARRAY_H
#define INCL_CPPTYPEARRAY_H

#include "CPPTypeModifier.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeArray type class.  This class holds declarations
// of unnamed types that are arrays of other types 
///////////////////////////////////////////////////////////////////////
class CPPTypeArray : public CPPTypeModifier
{
public:
   // ctor/dtor
   CPPTypeArray(
      CPPTypeRefI baseType_,
      int dimension_,
      bool addToBase
   );
   virtual ~CPPTypeArray();

   // access methods
   int GetDimension() const { assertValid(); return dimension; }

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Comparison of two array type nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return
         CPPTypeModifier::LessThanNode(rhs) ||
         (
            rhs.GetFlavor() == Array &&
            GetDimension() < STATIC_CAST(const CPPTypeArray*, &rhs)->GetDimension()
         );
   }

private:
   // disallow copy or assignment
   CPPTypeArray(const CPPTypeArray&);
   void operator=(const CPPTypeArray&);

   void assertValid() const { 
      CPPTypeModifier::assertValid(); 
      assert(GetFlavor() == Array); 
   }

   // dimension is zero for "[]" declarators
   int dimension;
};


#endif
