// CPPTypeBitfield.h: definition of bitfield-type modifier class

#ifndef INCL_CPPTYPEBITFIELD_H
#define INCL_CPPTYPEBITFIELD_H

#include "CPPTypeModifier.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeBitfield type class.  This class holds declarations
// of bitfield types that are derived from other types 
///////////////////////////////////////////////////////////////////////
class CPPTypeBitfield : public CPPTypeModifier
{
public:
   // ctor/dtor
   CPPTypeBitfield(
      CPPTypeRefI baseType_,
      int fieldlen_,
      bool addToBase
   );
   virtual ~CPPTypeBitfield();

   // access methods
   int GetFieldlen() const { assertValid(); return fieldlen; }

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Comparison of two bitfield type nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return
         CPPTypeModifier::LessThanNode(rhs) ||
         (
            rhs.GetFlavor() == Bitfield &&
            GetFieldlen() < STATIC_CAST(const CPPTypeBitfield*, &rhs)->GetFieldlen()
         );
   }

private:
   // disallow copy or assignment
   CPPTypeBitfield(const CPPTypeBitfield&);
   void operator=(const CPPTypeBitfield&);

   void assertValid() const { 
      CPPTypeModifier::assertValid(); 
      assert(GetFlavor() == Bitfield); 
   }

   int fieldlen;
};

#endif
