// CPPTypeBuiltin.h: definition of builtin-type class

#ifndef INCL_CPPTYPEBUILTIN_H
#define INCL_CPPTYPEBUILTIN_H

#include "CPPType.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeBuiltin declaration class.  This class holds declarations of
// combinations of various builtin types (int/long/double/unsigned/etc).
///////////////////////////////////////////////////////////////////////
class CPPTypeBuiltin : public CPPType
{
public:
   CPPTypeBuiltin(
      CPPConst::TypeSpecifier typeSpecifier_
   );
   virtual ~CPPTypeBuiltin() { assertValid(); }

   // access methods
   CPPConst::TypeSpecifier GetTypeSpecifier() const { 
      assertValid(); return typeSpecifier; 
   }

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Format the name of the type specifiers (may be used for invalid types)
   static JLStr FormatTypeSpecifierName(CPPConst::TypeSpecifier ts);

   // Method to validate builtin types
   static bool TypeSpecifierIsValid(CPPConst::TypeSpecifier typeSpecifier);

   // Ordering of two nodes
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return 
         CPPType::LessThanNode(rhs) ||
         (
            rhs.GetFlavor() == Builtin && 
            typeSpecifier < STATIC_CAST(const CPPTypeBuiltin*, &rhs)->GetTypeSpecifier()
         );
   }

private:
   // disallow copy or assignment
   CPPTypeBuiltin(const CPPTypeBuiltin&);
   void operator=(const CPPTypeBuiltin&);

   void assertValid() const { CPPType::assertValid(); assert(GetFlavor() == Builtin); }

   CPPConst::TypeSpecifier typeSpecifier;

   // Initialize the static data members if necessary
   static void InitTables();

	class CompatTableVector;
   static CompatTableVector *compatTable;

	class NameTableVector;
   static NameTableVector *nameTable;
};

#endif
