// CPPTypeCVQualified: Implementation for const/volatile type modifier class.

#include "CPPType.h"
#include "CPPTypeCVQualified.h"

/////////////////////////////////////////////////////////////////////////
// ctor/dtor
/////////////////////////////////////////////////////////////////////////
CPPTypeCVQualified::CPPTypeCVQualified(
   CPPTypeRefI baseType_,
   CPPConst::CVQualifier cvQualifier_,
   bool addToBase
) : 
   CPPTypeModifier(CVQualified, baseType_),
   cvQualifier(cvQualifier_)
{
   assert(baseType_.get() != 0);
   if (addToBase)
   {
      AddToBase();
   }
}

//virtual
CPPTypeCVQualified::~CPPTypeCVQualified()
{ 
   assertValid(); 
   RemoveFromBase(); 
}

/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                              to format the declarator name
//                              in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeCVQualified::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI,
   bool forSpecialization
) const
{
   assertValid();

   JLStr modifiedDeclarator;

   if ((cvQualifier & CPPConst::cvCONST) != 0)
   {
      modifiedDeclarator += "const ";
   }

   if ((cvQualifier & CPPConst::cvVOLATILE) != 0)
   {
      modifiedDeclarator += "volatile ";
   }

   modifiedDeclarator += declName;

   return GetBaseType()->FormatName(
      modifiedDeclarator, 
      suffixStr,
      StrListRefI(),
      forSpecialization
   );
}


/////////////////////////////////////////////////////////////////////
// StripCVQualifiers: Remove const/volatile modifiers from a type and
// return the resulting type.
//
// Inputs: none
// Outputs: none
// Return:
//    CPPTypeRefI   The type with any templated modifiers removed
/////////////////////////////////////////////////////////////////////
CPPTypeRefI
CPPTypeCVQualified::StripCVQualifiers()
{
   return GetBaseType()->StripCVQualifiers();
}

