// CPPTypeCVQualified.h: definition of const/volatile type-modifier class

#ifndef INCL_CPPTYPECVQUALIFIED_H
#define INCL_CPPTYPECVQUALIFIED_H

#include "CPPTypeModifier.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeCVQualified type class.  This class holds const- or
// volatile- qualifications of other types.
///////////////////////////////////////////////////////////////////////
class CPPTypeCVQualified : public CPPTypeModifier
{
public:
   CPPTypeCVQualified(
      CPPTypeRefI baseType_,
      CPPConst::CVQualifier cvQualifier_,
      bool addToBase
   );
   virtual ~CPPTypeCVQualified();

   // Remove CV qualifiers
   virtual CPPTypeRefI StripCVQualifiers();

   // Is this a function type?  const/volatile function is still a function
   virtual bool IsFunctionType() const { return GetBaseType()->IsFunctionType(); }

   // Find the return type of any function that this modifies
   virtual CPPTypeRefI GetFunctionReturnType() const { return GetBaseType()->GetFunctionReturnType(); }

   // access methods
   CPPConst::CVQualifier GetCVQualifier() const { 
      assertValid(); return cvQualifier; 
   }

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Comparison of two cv-qualified type nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return 
         CPPTypeModifier::LessThanNode(rhs) || 
         (
            rhs.GetFlavor() == CVQualified && 
            cvQualifier <  STATIC_CAST(const CPPTypeCVQualified*, &rhs)->GetCVQualifier()
         );
   }

private:
   // disallow copy or assignment
   CPPTypeCVQualified(const CPPTypeCVQualified&);
   void operator=(const CPPTypeCVQualified&);

   void assertValid() const { CPPTypeModifier::assertValid(); assert(flavor == CVQualified); }
   CPPConst::CVQualifier cvQualifier;
};


#endif
