// CPPTypeEnum.h: definition of enum-describing class

#ifndef INCL_CPPTYPEENUM_H
#define INCL_CPPTYPEENUM_H

#include "CPPType.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeEnum type class.  This class holds types for enumerations. 
//
// Note: For each *named* enumeration type, there will be an entry
// in the type name map for the containing scope, that associates
// the name with the enumeration type.  There will also be entries
// in the containing scope for each enumerator value, which show
// as special declarations of "constant enumerators" that have the
// enumeration as their type.
///////////////////////////////////////////////////////////////////////
class CPPTypeEnum : public CPPType
{
public:
   // ctor/dtor
   CPPTypeEnum(const JLStr& name_) : CPPType(Enum), name(name_), complete(false) {}
   virtual ~CPPTypeEnum() {}

   // access methods
   const JLStr& GetName() { assertValid(); return name; }
   virtual bool GetComplete() const { return complete; }
   virtual void SetComplete() { complete = true; }

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Equality of two enum type nodes.
   // Must be the same node since enums are unique
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return 
         CPPType::LessThanNode(rhs) || 
         rhs.GetFlavor() == Enum && this < &rhs;
   }

private:
   // disallow copy or assignment
   CPPTypeEnum(const CPPTypeEnum&);
   void operator=(const CPPTypeEnum&);

   void assertValid() const { CPPType::assertValid(); assert(GetFlavor() == Enum); }
   // Note that enumerator *values* are not part of the type.
   // They are separate declarations in the containing scope.

   JLStr name;
   bool complete;
};


#endif
