// CPPTypeIndirection.h: definition of pointer/reference type modifier class

#ifndef INCL_CPPTYPEINDIRECTION_H
#define INCL_CPPTYPEINDIRECTION_H

#include "CPPTypeModifier.h"

///////////////////////////////////////////////////////////////////////
// CPPTypeIndirection type class.  This class holds types
// that are indirections (pointer or reference) to other types.
///////////////////////////////////////////////////////////////////////
class CPPTypeIndirection : public CPPTypeModifier
{
public:
   enum IndirectType { 
      IndirectPointer, 
      IndirectReference
   };

   // ctor/dtor
   CPPTypeIndirection(
      CPPTypeRefI baseType_,
      IndirectType indirectType_,
      bool addToBase
   );
   virtual ~CPPTypeIndirection();

   // access methods
   IndirectType GetIndirectType() const { 
      assertValid(); return indirectType; 
   }

   // Remove reference
   virtual CPPTypeRefI StripReference();

   // Remove pointer
   virtual CPPTypeRefI StripPointer();

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Comparison of two indirection type nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return
         CPPTypeModifier::LessThanNode(rhs) ||
         (
            rhs.GetFlavor() == Indirection &&
            indirectType < STATIC_CAST(const CPPTypeIndirection*, &rhs)->GetIndirectType()
         );
   }

private:
   // disallow copy or assignment
   CPPTypeIndirection(const CPPTypeIndirection&);
   void operator=(const CPPTypeIndirection&);

   void assertValid() const { 
      CPPTypeModifier::assertValid(); 
      assert(flavor == Indirection); 
   }
   IndirectType indirectType;
};



#endif
