// CPPTypePointerToMember.h: definition of pointer-to-member type
// modifier class

#ifndef INCL_CPPTYPEPOINTERTOMEMBER_H
#define INCL_CPPTYPEPOINTERTOMEMBER_H

#include "CPPTypeModifier.h"

///////////////////////////////////////////////////////////////////////
// CPPTypePointerToMember type class.  This class holds types that
// are pointers-to-members-of-a-class.  These types modify the
// types they point to, but also have a "ptrToMemberClass", which
// is the class type whose member is pointed to.
// 
// The ptrToMemberClass type must be CPPTypeClass.
///////////////////////////////////////////////////////////////////////
class CPPTypePointerToMember : public CPPTypeModifier
{
public:
   // ctor/dtor
   CPPTypePointerToMember(
      CPPTypeRefI baseType_,
      CPPTypeRefI ptrToMemberClass_,
      bool addToBase
   );
   virtual ~CPPTypePointerToMember();

   // access methods
   CPPTypeRefI GetPtrToMemberClass() const { 
      assertValid(); 
      return ptrToMemberClass;
   }

   // Find a template placeholder type in the modifier chain
   virtual bool FindTemplatePlaceholder(CPPTypeRefI& typeReturn);

   // Format the name of the type
   virtual JLStr FormatName(
      const JLStr& declName, 
      const JLStr& suffixStr,
      StrListRefI argList,
      bool forSpecialization
   ) const;

   // Comparison of two pointer-to type nodes.
   virtual bool LessThanNode(const CPPType& rhs) const
   {
      assertValid();
      return
         CPPTypeModifier::LessThanNode(rhs) ||
         (
            rhs.GetFlavor() == PointerToMember &&
            GetPtrToMemberClass() < 
               STATIC_CAST(const CPPTypePointerToMember*, &rhs)->GetPtrToMemberClass()
         );
   }

private:
   // disallow copy or assignment
   CPPTypePointerToMember(const CPPTypePointerToMember&);
   void operator=(const CPPTypePointerToMember&);

   void assertValid() const { 
      CPPTypeModifier::assertValid(); 
      assert(GetFlavor() == PointerToMember); 
   }

   // This can be a smart pointer because types do not own 
   // their modifiers; modifiers own the types.
   CPPTypeRefI ptrToMemberClass;
};

#endif
