// CPPTypeTemplatePlaceholder.cpp.  Definition of template-type-parameter placeholder type.

#include "CPPTypeTemplatePlaceholder.h"
#include "CPPScopeTemplate.h"

/////////////////////////////////////////////////////////////////
// Class statics
/////////////////////////////////////////////////////////////////

// static template placeholder scope.  This is represents any
// and all synthetic scopes used by template placeholders.  When
// a scope search is done and it finds the name of a template
// placeholder type, then this scope is always returned.
CPPScopeRefI CPPTypeTemplatePlaceholder::thePlaceholderScope = new CPPScopeTemplate();

/////////////////////////////////////////////////////////////////
// FindTemplatePlaceholder: Search the type chain for a template
// placeholder.  Of course, a template placeholder returns itself.
//
// Inputs: none
// Outputs:
//    CPPTypeRefI&   typeReturn   This type.
// Return:
//    bool     true
/////////////////////////////////////////////////////////////////
//virtual 
bool 
CPPTypeTemplatePlaceholder::FindTemplatePlaceholder(
   CPPTypeRefI& typeReturn
)
{
   typeReturn = this;
   return true;
}

/////////////////////////////////////////////////////////////////
// FormatName: Format the name of the type, using the names of the
// modified types where needed to construct the full type name.
//
// Inputs:
//    const JLStr&  declName   If non-empty, then this is used
//                             to format the declarator name
//                             in the proper position.
//    const JLStr&  suffixStr  Any accumulated suffix string
//    bool          forSpecialization
//                             If true, then the name is formatted
//                             to be suitable in constructing a
//                             specialization scope.
// Return:
//    JLStr    The formatted type name
/////////////////////////////////////////////////////////////////
// virtual 
JLStr 
CPPTypeTemplatePlaceholder::FormatName(
   const JLStr& declName,
   const JLStr& suffixStr,
   StrListRefI,
   bool forSpecialization
) const
{
   assertValid();

   char buf[10];
   sprintf(buf, "%d", parameterIdx);
   char buf2[10];
   if (templateLevel != 0)
   {
      sprintf(buf2, "_%d", templateLevel);
   } else {
      buf2[0] = 0;
   }
   JLStr retval = JLStr("$") + buf + buf2;

   if (declName.length() != 0 || suffixStr.length() != 0)
   {
      JLStr tmpStr = declName + suffixStr;
      if (isalnum(tmpStr[0]))
      {
         retval += " ";
      }
      retval += tmpStr;
   }

   if (!forSpecialization)
   {
      retval = "typename " + retval;
   }
   return retval;
}

