// TemplateArgumentList.h: definition of TemplateArgument and TemplateArgumentList types

#ifndef INCL_TEMPLATEARGUMENTLIST_H
#define INCL_TEMPLATEARGUMENTLIST_H

#include "TemplateArgumentListIncomplete.h"
#include "CPPTypeIncomplete.h"
#include "CPPExprIncomplete.h"
#include "JLStr.h"
#include "JL_DList.h"

////////////////////////////////////////////////////////////////////////
// TemplateArgument: base class for the template argument types.  This is
// used to describe a single template argument in a template declaration.
////////////////////////////////////////////////////////////////////////
class TemplateArgument : public VRefCount
{
public:
   enum Flavor { Type, Value };
   TemplateArgument() {}
   virtual ~TemplateArgument() {}

   // Accessors
   virtual Flavor GetFlavor() const = 0;
   virtual CPPTypeRefI GetType() const { assert(0); return CPPTypeRefI(); }
   virtual CPPExprRefI GetValue() const { assert(0); return CPPExprRefI(); }
   virtual JLStr Format() const = 0;
};

////////////////////////////////////////////////////////////////////////
// TemplateArgumentType:  The TemplateArgument subclass used to describe template
// arguments that are types (as opposed to values)
////////////////////////////////////////////////////////////////////////
class TemplateArgumentType : public TemplateArgument
{
public:
   TemplateArgumentType(
      CPPTypeRefI type_
   ) :
      TemplateArgument(),
      type(type_)
   {}
   virtual ~TemplateArgumentType() {}

   // Accessors
   virtual Flavor GetFlavor() const { return Type; }
   virtual CPPTypeRefI GetType() const { return type; }
   virtual JLStr Format() const;

private:
   CPPTypeRefI type;
};

////////////////////////////////////////////////////////////////////////
// TemplateArgumentValue:  The TemplateArgument subclass used to 
// describe template arguments that are values (as opposed to types)
////////////////////////////////////////////////////////////////////////
class TemplateArgumentValue : public TemplateArgument
{
public:
   TemplateArgumentValue(
      CPPExprRefI value_
   ) : 
      TemplateArgument(),
      value(value_)
   {}
   virtual ~TemplateArgumentValue() {}

   // Accessors
   virtual Flavor GetFlavor() const { return Value; }
   virtual CPPExprRefI GetValue() const { return value; }
   virtual JLStr Format() const;

private:
   CPPExprRefI value;
};

////////////////////////////////////////////////////////////////////////
// TemplateArgumentList:  List of template arguments
////////////////////////////////////////////////////////////////////////
class TemplateArgumentList : 
   public VRefCount,
   public JL_DList<TemplateArgumentRefI, allocator<TemplateArgumentRefI> > 
{
};

#endif
