// TemplateParameterList.h: definition of TemplateParameter and TemplateParameterList types

#ifndef INCL_TEMPLATEPARAMETERLIST_H
#define INCL_TEMPLATEPARAMETERLIST_H

#include "TemplateParameterListIncomplete.h"
#include "CPPTypeIncomplete.h"
#include "CPPExprIncomplete.h"
#include "TokenListIncomplete.h"
#include "JLStr.h"
#include "JL_DList.h"

////////////////////////////////////////////////////////////////////////
// TemplateParameter: base class for the template parameter types.  This is
// used to describe a single template parameter in a template declaration.
////////////////////////////////////////////////////////////////////////
class TemplateParameter : public VRefCount
{
public:
   enum Flavor { Type, Value };
   TemplateParameter(
      const JLStr& name_,
      TokenListRefI defaultParameterTokens_
   );
   virtual ~TemplateParameter();

   // Accessors
   virtual Flavor GetFlavor() const = 0;
   virtual CPPTypeRefI GetType() const = 0;
   virtual JLStr Format() const = 0;
   const JLStr& GetName() const { return name; }
   TokenListRefI GetDefaultParameterTokens() const { return defaultParameterTokens; }

private:
   JLStr name;
   TokenListRefI defaultParameterTokens;
};

////////////////////////////////////////////////////////////////////////
// TemplateParameterType:  The TemplateParameter subclass used to describe 
// template parameters that are types (as opposed to values)
////////////////////////////////////////////////////////////////////////
class TemplateParameterType : public TemplateParameter
{
public:
   TemplateParameterType(
      const JLStr& name_,
      TokenListRefI defaultParameterTokens_,
      CPPTypeRefI placeholderType_
   );
   virtual ~TemplateParameterType() {}

   // Accessors
   virtual Flavor GetFlavor() const { return Type; }
   virtual CPPTypeRefI GetType() const { return placeholderType; }
   virtual JLStr Format() const;
private:
   CPPTypeRefI placeholderType;
};

////////////////////////////////////////////////////////////////////////
// TemplateParameterValue:  The TemplateParameter subclass used to 
// describe template parameters that are values (as opposed to types)
////////////////////////////////////////////////////////////////////////
class TemplateParameterValue : public TemplateParameter
{
public:
   TemplateParameterValue(
      const JLStr& name_,
      TokenListRefI defaultParameterTokens_,
      CPPTypeRefI parameterType_,
      CPPExprRefI placeholderValue_
   ) : 
      TemplateParameter(name_, defaultParameterTokens_),
      parameterType(parameterType_),
      placeholderValue(placeholderValue_)
   {}
   virtual ~TemplateParameterValue() {}

   // Accessors
   virtual Flavor GetFlavor() const { return Value; }
   virtual CPPTypeRefI GetType() const { return parameterType; }
   virtual JLStr Format() const;
   CPPExprRefI GetValue() const { return placeholderValue; }

private:
   CPPTypeRefI parameterType;
   CPPExprRefI placeholderValue;
};

////////////////////////////////////////////////////////////////////////
// TemplateParameterList:  List of ref-counted template parameters
////////////////////////////////////////////////////////////////////////
class TemplateParameterList : 
   public VRefCount, 
   public JL_DList<TemplateParameterRefI, allocator<TemplateParameterRefI> >
{
};

#endif
