#include "CPreToCPPBuffer.h"
#include "BufferedCPreParser.h"
#include "FastToken.h"

CPreToCPPBuffer::CPreToCPPBuffer(
   BufferedCPreParser *preprocessor_
) :
   preprocessor(preprocessor_)
{
}


// CPreToCPPBuffer::AddToken.  Add a token (presumably from the
// preprocessor) to the buffer.
void
CPreToCPPBuffer::AddToken(
   const ANTLRTokenPtr& token
)
{
   tokens.append(token);
}

// CPreToCPPBuffer::MoreTokens.  Get more tokens from
// the preprocessor object and put them into the buffer
// until we have at least the desired number of buffered tokens.
void
CPreToCPPBuffer::MoreTokens()
{
   while (tokens.entries() == 0) {
      if (preprocessor->preprocessing_chunk())
      {
         // Preprocessor has not reached Eof, and may have added tokens
         // to the buffer.
      } else {
         // End-of-file reached -- create and add an Eof token
         tokens.append(ANTLRTokenPtr(FastToken::newToken(TOK_Eof)));
      }
   }
}


// CPreToCPPBuffer::getToken.  Get the next token in the buffer.
ANTLRTokenPtr 
CPreToCPPBuffer::getToken()
{
   if (tokens.entries() == 0) {
      MoreTokens();
   }

   return tokens.next();
}

