// CPreToCPPBuffer.h:  Token buffer object for the preprocessor.
// This token buffer object was created to support a stack of token
// buffers, which in turn support the processing of #include directives.
// The token buffer must be able to:
//    1) Allow the "pushing" of another input token stream onto a
//       stack of input tokens.
//    2) Automatically popping the topmost input stream when
//       Eof is returned from the input token stream.
//
// Although this is a subclass of ANTLRTokenBuffer, it uses nothing
// from the base class except the interface.  It is unfortunate that
// no abstract base class was defined for this purpose, since the
// additional function of ANTLRTokenBuffer is superfluous.

#ifndef INCL_CPRETOCPPBUFFER_H
#define INCL_CPRETOCPPBUFFER_H

#include "../cpre/tokens.h"
#include "ATokenStream.h"
#include "ATokPtr.h"
#include "assert.h"
#include "JLCircBuffer.h"
#include "BufferedCPreParser.h"

class CPreToCPPBuffer : public ANTLRTokenStream
{
public:
   // ctor/dtor
   CPreToCPPBuffer(BufferedCPreParser *preprocessor_);
   virtual ~CPreToCPPBuffer() {}

   virtual ANTLRTokenPtr getToken();

   // Token buffer interface from preprocessor
   void AddToken(const ANTLRTokenPtr& token);

private:
   // Preprocessor object
   BufferedCPreParser *preprocessor;

   // Called to get more tokens placed into the token buffer
   void MoreTokens();
   
   // token buffer
   JLCircBuffer<ANTLRTokenPtr> tokens;
};

#endif
