// JLTokenBuffer.h:  Token buffer object.

#ifndef INCL_JLTOKENBUFFER_H
#define INCL_JLTOKENBUFFER_H

#include "STLIncomplete.h"
#include "tokens.h"
#include "ATokenBuffer.h"
#include "ATokPtr.h"
#include "assert.h"
#include "JLCircBuffer.h"

class ANTLRTokenStream;

class JLTokenBuffer : public ANTLRTokenBuffer
{
public:
   // ctor/dtor
   JLTokenBuffer(ANTLRTokenStream *input_, int k_);
   virtual ~JLTokenBuffer();

   // Token buffer interface to parser
   virtual void consume();
   virtual ANTLRTokenPtr LT(int i);
   virtual int mark();
   virtual void rewind(int pos);
   virtual void unmark(int pos);

private:
   // Fill up the token buffer
   void FillBuffer(int amount);

   // Input token stream
   ANTLRTokenStream *input;

   // Lookahead amount
   int k;
   
   // number of active markers
   int num_markers;

   // additional offset into token buffer used when mark is set (guess mode)
   int offset;

   // token circular buffer
   JLCircBuffer<ANTLRTokenPtr> tokens;

   // magic number and method to catch any incorrect downcasting
   #ifdef _DEBUG
      enum { Magic = 0x214E };
      short magic;
      void validate() const { assert(magic == Magic); }
   #else
      void validate() const {}
   #endif
};

#endif
