/*
 * PUBLIC DOMAIN PCCTS-BASED C++ PREPROCESSOR GRAMMAR
 *
 * This is a companion preprocessor designed to work closely with
 * the PCCTS-based C++ grammar.  See the README for more information.
 *
 * Preprocessor Grammar:
 * Author:  
 *       John Lilley, Empathy Software     
 *       jlilley@empathy.com
 *       http://www.empathy.com
 *
 * Portions of preprocessor derived from the corresponding
 * public domain C++ grammar courtesy of the following people:
 *      Sumana Srinivasan, NeXT Inc.;               sumana_srinivasan@next.com
 *      Terence Parr, Parr Research Corporation;    parrt@parr-research.com
 *      Russell Quong, Purdue University;           quong@ecn.purdue.edu
 *
 * VERSION 1.0
 *
 * SOFTWARE RIGHTS
 * These software rights apply to the portions of this public-domain
 * C++ preprocessor grammar developed by either party listed above.
 * 
 * This file is a part of the ANTLR-based C++ grammar and is free
 * software.  We do not reserve any LEGAL rights to its use or
 * distribution, but you may NOT claim ownership or authorship of this
 * grammar or support code.  An individual or company may otherwise do
 * whatever they wish with the grammar distributed herewith including the
 * incorporation of the grammar or the output generated by ANTLR into
 * commerical software.  You may redistribute in source or binary form
 * without payment of royalties to us as long as this header remains
 * in all source distributions.
 *
 * We encourage users to develop parsers/tools using this grammar.
 * In return, we ask that credit is given to us for developing this
 * grammar.  By "credit", we mean that if you incorporate our grammar or
 * the generated code into one of your programs (commercial product,
 * research project, or otherwise) that you acknowledge this fact in the
 * documentation, research report, etc....  In addition, you should say nice
 * things about us at every opportunity.
 *
 * As long as these guidelines are kept, we expect to continue enhancing
 * this grammar.  Feel free to send us enhancements, fixes, bug reports,
 * suggestions, or general words of encouragement at parrt@parr-research.com.
 *
 * C++ Preprocessor questions and concerns should be sent to:
 *      jlilley@empathy.com
 * 
 * C++ Grammar conceived in conjunction with a project at:
 *      NeXT Computer Inc.
 *      900 Chesapeake Dr.
 *      Redwood City, CA 94555
 * 
 * Built and tested with PCCTS version 1.33
 * 
 * DISCLAIMER: we make no guarantees that this grammar works, makes sense,
 * or can be used to do anything useful.
 * [ John Lilley ] ... but if it is broken, I or others can help on
 * mercenary terms  :-)
 *
 */

#header <<
#ifdef _MSC_VER
   // Disable some warnings that appear commonly in Rogue Wave STL
   #pragma warning (disable:4786)    // long debug names
   #pragma warning (disable:4146)    // unary minus operator applied to unsigned type
   #pragma warning (disable:4018)    // '>' : signed/unsigned mismatch
#endif
#include "tokens.h"
#include "FastToken.h"
#include "ATokPtr.h"
#include "JLStr.h"
class TokenList;
>>

// In order for the preprocessor to pass the correct tokens to the parser,
// these token ids should be used in the C++ parser instead of string literals.
// In addition, the parser should use #tokdef with the tokens.h created from
// the preprocessor, instead of using its own token definitions.

// Note:  This grammar requires a "modified" lexer that has a flag
// "firstLineToken" in the class, initialized to true (1).  This can
// be done by modifiying "output.c" in the DLG source (preferred), or
// by editing the resulting DLG output.

// ANSI C++ Operator and punctuator tokens
#token TOK_AMPERSAND            "&"        <<firstLineToken=0;>>
#token TOK_AND                  "&&"       <<firstLineToken=0;>>
#token TOK_ASSIGNEQUAL          "="        <<firstLineToken=0;>>
#token TOK_BITWISEANDEQUAL      "&="       <<firstLineToken=0;>>
#token TOK_BITWISEOR            "\|"       <<firstLineToken=0;>>
#token TOK_BITWISEOREQUAL       "\|="      <<firstLineToken=0;>>
#token TOK_BITWISEXOR           "^"        <<firstLineToken=0;>>
#token TOK_BITWISEXOREQUAL      "^="       <<firstLineToken=0;>>
#token TOK_COMMA                ","        <<firstLineToken=0;>>
#token TOK_COLON                ":"        <<firstLineToken=0;>>
#token TOK_DIVIDE               "/"        <<firstLineToken=0;>>
#token TOK_DIVIDEEQUAL          "/="       <<firstLineToken=0;>>
#token TOK_DOT                  "."        <<firstLineToken=0;>>
#token TOK_DOTMBR               ".\*"      <<firstLineToken=0;>>
#token TOK_ELLIPSIS             "..."      <<firstLineToken=0;>>
#token TOK_EQUAL                "=="       <<firstLineToken=0;>>
#token TOK_GREATERTHAN          ">"        <<firstLineToken=0;>>
#token TOK_GREATERTHANOREQUALTO ">="       <<firstLineToken=0;>>
#token TOK_LCURLY               "\{"       <<firstLineToken=0;>>
#token TOK_LESSTHAN             "<"        <<firstLineToken=0;>>
#token TOK_LESSTHANOREQUALTO    "<="       <<firstLineToken=0;>>
#token TOK_LPAREN               "\("       <<firstLineToken=0;>>
#token TOK_LSQUARE              "\["       <<firstLineToken=0;>>
#token TOK_MINUS                "\-"       <<firstLineToken=0;>>
#token TOK_MINUSEQUAL           "\-="      <<firstLineToken=0;>>
#token TOK_MINUSMINUS           "\-\-"     <<firstLineToken=0;>>
#token TOK_MOD                  "%"        <<firstLineToken=0;>>
#token TOK_MODEQUAL             "%="       <<firstLineToken=0;>>
#token TOK_NOT                  "!"        <<firstLineToken=0;>>
#token TOK_NOTEQUAL             "!="       <<firstLineToken=0;>>
#token TOK_OR                   "\|\|"     <<firstLineToken=0;>>
#token TOK_PLUS                 "\+"       <<firstLineToken=0;>>
#token TOK_PLUSEQUAL            "\+="      <<firstLineToken=0;>>
#token TOK_PLUSPLUS             "\+\+"     <<firstLineToken=0;>>
#token TOK_POINTERTO            "\->"      <<firstLineToken=0;>>
#token TOK_POINTERTOMBR         "\->\*"    <<firstLineToken=0;>>
#token TOK_QUESTIONMARK         "?"        <<firstLineToken=0;>>
#token TOK_RCURLY               "\}"       <<firstLineToken=0;>>
#token TOK_RPAREN               "\)"       <<firstLineToken=0;>>
#token TOK_RSQUARE              "\]"       <<firstLineToken=0;>>
#token TOK_SCOPE                "::"       <<firstLineToken=0;>>
#token TOK_SEMICOLON            ";"        <<firstLineToken=0;>>
#token TOK_SHIFTLEFT            "\<\<"     <<firstLineToken=0;>>
#token TOK_SHIFTLEFTEQUAL       "\<\<="    <<firstLineToken=0;>>
#token TOK_SHIFTRIGHT           "\>\>"     <<firstLineToken=0;>>
#token TOK_SHIFTRIGHTEQUAL      "\>\>="    <<firstLineToken=0;>>
#token TOK_STAR                 "\*"       <<firstLineToken=0;>>
#token TOK_TILDE                "\~"       <<firstLineToken=0;>>
#token TOK_TIMESEQUAL           "\*="      <<firstLineToken=0;>>

// ANSI C++ Keywords
#token TOK_ASM              "asm"              <<firstLineToken=0;>>
#token TOK_AUTO             "auto"             <<firstLineToken=0;>>
#token TOK_BOOL             "bool"             <<firstLineToken=0;>>
#token TOK_BOOLFALSE        "false"            <<firstLineToken=0;>>
#token TOK_BOOLTRUE         "true"             <<firstLineToken=0;>>
#token TOK_BREAK            "break"            <<firstLineToken=0;>>
#token TOK_CASE             "case"             <<firstLineToken=0;>>
#token TOK_CATCH            "catch"            <<firstLineToken=0;>>
#token TOK_CHAR             "char"             <<firstLineToken=0;>>
#token TOK_CLASS            "class"            <<firstLineToken=0;>>
#token TOK_CONST            "const"            <<firstLineToken=0;>>
#token TOK_CONST_CAST       "const_cast"       <<firstLineToken=0;>>
#token TOK_CONTINUE         "continue"         <<firstLineToken=0;>>
#token TOK_DEFAULT          "default"          <<firstLineToken=0;>>
#token TOK_DEFINED          "defined"          <<firstLineToken=0;>>
#token TOK_DELETE           "delete"           <<firstLineToken=0;>>
#token TOK_DO               "do"               <<firstLineToken=0;>>
#token TOK_DOUBLE           "double"           <<firstLineToken=0;>>
#token TOK_DYNAMIC_CAST     "dynamic_cast"     <<firstLineToken=0;>>
#token TOK_ELSE             "else"             <<firstLineToken=0;>>
#token TOK_ENUM             "enum"             <<firstLineToken=0;>>
#token TOK_EXPLICIT         "explicit"         <<firstLineToken=0;>>
#token TOK_EXTERN           "extern"           <<firstLineToken=0;>>
#token TOK_FLOAT            "float"            <<firstLineToken=0;>>
#token TOK_FOR              "for"              <<firstLineToken=0;>>
#token TOK_FRIEND           "friend"           <<firstLineToken=0;>>
#token TOK_GOTO             "goto"             <<firstLineToken=0;>>
#token TOK_IF               "if"               <<firstLineToken=0;>>
#token TOK_INLINE           "inline"           <<firstLineToken=0;>>
#token TOK_INT              "int"              <<firstLineToken=0;>>
#token TOK_LONG             "long"             <<firstLineToken=0;>>
#token TOK_MUTABLE          "mutable"          <<firstLineToken=0;>>
#token TOK_NAMESPACE        "namespace"        <<firstLineToken=0;>>
#token TOK_NEW              "new"              <<firstLineToken=0;>>
#token TOK_OPERATOR         "operator"         <<firstLineToken=0;>>
#token TOK_PRIVATE          "private"          <<firstLineToken=0;>>
#token TOK_PROTECTED        "protected"        <<firstLineToken=0;>>
#token TOK_PUBLIC           "public"           <<firstLineToken=0;>>
#token TOK_REGISTER         "register"         <<firstLineToken=0;>>
#token TOK_REINTERPRET_CAST "reinterpret_cast" <<firstLineToken=0;>>
#token TOK_RETURN           "return"           <<firstLineToken=0;>>
#token TOK_SHORT            "short"            <<firstLineToken=0;>>
#token TOK_SIGNED           "signed"           <<firstLineToken=0;>>
#token TOK_SIZEOF           "sizeof"           <<firstLineToken=0;>>
#token TOK_STATIC           "static"           <<firstLineToken=0;>>
#token TOK_STATIC_CAST      "static_cast"      <<firstLineToken=0;>>
#token TOK_STRUCT           "struct"           <<firstLineToken=0;>>
#token TOK_SWITCH           "switch"           <<firstLineToken=0;>>
#token TOK_TEMPLATE         "template"         <<firstLineToken=0;>>
#token TOK_THIS             "this"             <<firstLineToken=0;>>
#token TOK_THROW            "throw"            <<firstLineToken=0;>>
#token TOK_TRY              "try"              <<firstLineToken=0;>>
#token TOK_TYPEDEF          "typedef"          <<firstLineToken=0;>>
#token TOK_TYPEID           "typeid"           <<firstLineToken=0;>>
#token TOK_TYPENAME         "typename"         <<firstLineToken=0;>>
#token TOK_UNION            "union"            <<firstLineToken=0;>>
#token TOK_UNSIGNED         "unsigned"         <<firstLineToken=0;>>
#token TOK_USING            "using"            <<firstLineToken=0;>>
#token TOK_VIRTUAL          "virtual"          <<firstLineToken=0;>>
#token TOK_VOID             "void"             <<firstLineToken=0;>>
#token TOK_VOLATILE         "volatile"         <<firstLineToken=0;>>
#token TOK_WCHAR_T          "wchar_t"          <<firstLineToken=0;>>
#token TOK_WHILE            "while"            <<firstLineToken=0;>>

///////////////////////////////////////////////////////////////////////////
// Microsoft C++ specific language extension tokens
// Note: These definitions must preceed TOK_ID
///////////////////////////////////////////////////////////////////////////
#token TOK_MSEXT_INT8       "__int8"           <<firstLineToken=0;>>
#token TOK_MSEXT_INT16      "__int16"          <<firstLineToken=0;>>
#token TOK_MSEXT_INT32      "__int32"          <<firstLineToken=0;>>
#token TOK_MSEXT_INT64      "__int64"          <<firstLineToken=0;>>
#token TOK_MSEXT_BASED      "{_} _based"       <<firstLineToken=0;>>
#token TOK_MSEXT_DECLSPEC   "{_} _declspec"    <<firstLineToken=0;>>
#token TOK_MSEXT_CDECL      "{_} _cdecl"       <<firstLineToken=0;>>
#token TOK_MSEXT_FASTCALL   "{_} _fastcall"    <<firstLineToken=0;>>
#token TOK_MSEXT_STDCALL    "{_} _stdcall"     <<firstLineToken=0;>>
#token TOK_MSEXT_TRY        "__try"            <<firstLineToken=0;>>
#token TOK_MSEXT_EXCEPT     "__except"         <<firstLineToken=0;>>
#token TOK_MSEXT_FINALLY    "__finally"        <<firstLineToken=0;>>
#token TOK_MSEXT_LEAVE      "__leave"          <<firstLineToken=0;>>
#token TOK_MSEXT_INLINE     "{_} _inline"      <<firstLineToken=0;>>
#token TOK_MSEXT_ASM        "{_} _asm"         <<firstLineToken=0;>>


///////////////////////////////////////////////////////////////////////////
// Complex tokens
//
///////////////////////////////////////////////////////////////////////////
#token TOK_ID "[a-zA-Z_][a-zA-Z0-9_]*"                          <<firstLineToken=0;>>

#token TOK_OCTALINT "0[0-7]*{[uUlL]}"                           <<firstLineToken=0;>>

#token TOK_DECIMALINT "[1-9][0-9]*{[uUlL]}"                     <<firstLineToken=0;>>

#token TOK_HEXADECIMALINT "(0x|0X)[0-9a-fA-F]+{[uUlL]}"         <<firstLineToken=0;>>

#token TOK_FLOATONE 
       "([0-9]+ . [0-9]* | [0-9]* . [0-9]+) {[eE]{[\-\+]}[0-9]+} {[fFlL]}" 
       <<firstLineToken=0;>>

#token TOK_FLOATTWO "[0-9]+ [eE] {[\-\+]} [0-9]+ {[fFlL]}"      <<firstLineToken=0;>>

#token "/\*"                                     << mode (MLCOMMENT); skip ();>>

#token "\""                                      << mode(STRINGS); more (); >>

#token "'"                                       << mode(CHARACTERS); more (); >>

// Special End of file token
#token TOK_Eof "@"

///////////////////////////////////////////////////////////////////////////
// Preprocessor tokens
//
///////////////////////////////////////////////////////////////////////////

// backslash at end of line is skipped -- no token generated, 
// does not change firstLineToken status
#token "\\\n"                     << newline(); skip(); >>

#token "[\t\ ]+"                  << skip(); >>     // normal whitespace
#token "\12"                      << skip(); >>     // other whitespace

// Single-line comment tokens are generated, but stripped on output.
// The only reason we do not skip here, is that MFC has some nasty code
// that requires a macro-expanded comment to obliterate to end of line:
//         #define _WPAD /##/
//         _WPAD field;
#token TOK_SLCOMMENT "// ~[\n]*"      <<replchar(0);>>   // single-line comment (blank text)

// "Stringize" operator
#token TOK_PNDPND "##"                <<firstLineToken=0;>>

// TOK_PND may occur anywhere, but has special meaning at beginning of line
#token TOK_PND "#"  << if (firstLineToken) { more(); mode(DIRECTIVE_MODE); } >>

// newline token may enter other lexical classes
#token TOK_NEWLINE "\n"              << newline(); firstLineToken=1; >>

// Preprocessor tokens needed for <...> of #include, to allow
// for example #include <file.1xx> which would  otherwise have
// invalid tokens, and to prevent "1xx" being parsed as "1" "xx"
// This is ambiguous with many other tokens, and should come last
#token TOK_PP_NUMBER "[0-9] [0-9a-zA-Z_]+"      <<firstLineToken=0;>>

///////////////////////////////////////////////////////////////////////////
// Auxilliary lexical classes
//
///////////////////////////////////////////////////////////////////////////

// Special lexical class for recognizing directives at the beginning of the line
#lexclass DIRECTIVE_MODE
#token "[\t\ ]+"                   << skip(); >>
#token TOK_PNDDEFINE "define"          <<mode(DEFINING); firstLineToken=0;>>
#token TOK_PNDIF "if"                  <<mode(START); firstLineToken=0;>>
#token TOK_PNDIFDEF "ifdef"            <<mode(START); firstLineToken=0;>>
#token TOK_PNDIFNDEF "ifndef"          <<mode(START); firstLineToken=0;>>
#token TOK_PNDELIF "elif"              <<mode(START); firstLineToken=0;>>
#token TOK_PNDELSE "else"              <<mode(START); firstLineToken=0;>>
#token TOK_PNDENDIF "endif"            <<mode(START); firstLineToken=0;>>
#token TOK_PNDERROR "error"            <<mode(START); firstLineToken=0;>>
#token TOK_PNDINCLUDE "include"        <<mode(START); firstLineToken=0;>>
#token TOK_PNDLINE "line"              <<mode(START); firstLineToken=0;>>
#token TOK_PNDPRAGMA "pragma"          <<mode(START); firstLineToken=0;>>
#token TOK_PNDUNDEF "undef"            <<mode(START); firstLineToken=0;>>

// Special preprocessor-synthesized token denotes end-of-pragma
#token TOK_PRAGMA_END

//
// Special lexical class to distinguish function macro:
//         #define foo(x)
// from simple macro
//         #define foo (x)
//
#lexclass DEFINING
#token "[\t\ ]+"                                << skip(); >>
#token TOK_ID "[a-zA-Z_][a-zA-Z0-9_]*"              <<mode(START); firstLineToken=0;>>
#token TOK_FUNCID "[a-zA-Z_][a-zA-Z0-9_]* \("       <<mode(START); firstLineToken=0;>>

//
// Lexical class for multi-line comments
//
#lexclass MLCOMMENT
#token "[\n\r]"      << skip(); newline(); >>
#token "\*/"         << mode (START); skip (); firstLineToken=0; >>
#token "\*"          << skip (); >>
#token "~[\*\n\r]+"  << skip (); >>

//
// Lexical class for strings
// JEL removed escape processing from string tokens and character tokens,
// because they screw up include directives like:
//      #include "dir\file.h"
// or worse
//      #define file "dir\file.h"
//      #include file
// and they also didn't process escaped numerics "\012" correctly
//
#lexclass STRINGS
#token TOK_STRING "\""                  << mode (START); firstLineToken=0; >>
#token " \\ ( [ntvbrfa\\?'\"] | [0-9]+ | (x|X)[0-9a-fA-F]+ ) "  << more(); >>
#token "\\ [\n\r]"                  << newline(); more(); >>
#token "~[\\ \" \n \r]+"            << more(); >>


//
// Lexical class for character constants.
// Note that character constants may be "wide" like 'ABCD'
//
#lexclass CHARACTERS
#token TOK_CHARACTER "'"                << mode (START); firstLineToken=0; >>
#token " \\ ( [ntvbrfa\\?'\"] | [0-9]+ | (x|X)[0-9a-fA-F]+ ) "  << more(); >>
#token "~[\\ ' \n \r]+"                 << more(); >>


#lexclass START

///////////////////////////////////////////////////////////////////////////
// Grammar class declaration
//
///////////////////////////////////////////////////////////////////////////

// Note: most virtual methods in this class are there to support
// notification of other objects when certain actions occur.
class CPreParser {
<<
public:
   // need virtual dtor in case parser destroyed by ptr to base
   virtual void init();
   virtual ~CPreParser();

   // notification of include EOF from external buffer object
   virtual void OnIncludeEof() = 0;

   // Preprocessor options
   enum Option {
      OptMSExtensions,    // emit MS-extension tokens; otherwise they are IDs
      OptBool,            // emit bool,true,false tokens; otherwise they are IDs
      OptWCharT,          // emit wchar_t as a TOK_WCHAR_T token; otherwise it is an TOK_ID
      OptPragmas,         // send #pragma lines on to the C++ parser
      OptExpandPragmas,   // expand #pragma token list
      OptTrackInclude,    // track include stack on stdout
      OptTypename,        // emit 'typename' as TOK_TYPENAME, otherwise TOK_ID
      OptMax              // always last -- for error checking
   };

protected:
   //
   // Set/Get preprocessor options
   //
   virtual void SetOption(Option option, bool value) = 0;
   virtual bool GetOption(Option option) const = 0;

   //
   // Token lists for collecting tokens and identifiers
   // 
   TokenList* tempTokenList;
   TokenList* tempIdentifierList;
   virtual void ClearTokenList(TokenList* tokenList) = 0;
   virtual void AppendToTokenList(TokenList* tokenList, const ANTLRTokenPtr& tokPtr) = 0;

   // 
   // Symbol table management
   //
   // Implicitly uses tempTokenList
   virtual void EnterSimpleMacro(const ANTLRTokenPtr& name) = 0;

   // Implicitly uses tempIdentifierList and tempTokenList
   virtual void EnterFunctionMacro(const ANTLRTokenPtr& name) = 0;

   virtual void UndefineMacro(const ANTLRTokenPtr& name) = 0;
   virtual bool MacroIsDefined(const ANTLRTokenPtr& name) = 0;
   virtual bool IsFunctionLikeMacro(const ANTLRTokenPtr& name) = 0;

   //
   // #if block tracking.
   //
   virtual void EnterIfBlock(bool isTrue, const ANTLRTokenPtr& token) = 0;
   virtual void EnterElifBlock(bool isTrue, const ANTLRTokenPtr& token) = 0;
   virtual void EnterElseBlock(const ANTLRTokenPtr& token) = 0;
   virtual void ExitIfBlock(const ANTLRTokenPtr& token) = 0;
   virtual bool IfBlockTrue() const = 0;
   virtual bool EnclosingIfBlockTrue() const = 0;

   //
   // Evaluates the expression represented by tempTokenList
   virtual bool EvaluateExpression() = 0;

   //
   // File inclusion
   virtual void IncludeFile(const ANTLRTokenPtr& includeToken) = 0;

   // Methods for errors detected in code
   virtual void HandlePndError() = 0;
   virtual void HandleError(const JLStr& msg, const ANTLRTokenPtr& token) = 0;
   virtual void FatalError(const JLStr& msg) = 0;

   // Always call this method to output tokens
   // This method checks the token for type-conversion (bool, wchar_t, etc).
   virtual void OutputTokenBase(ANTLRTokenPtr& tokPtr) const = 0;

   // Output of tokens -- override these methods for physical token output
   virtual void ExpandAndOutputTokenList(TokenList *list) = 0;
   virtual void OutputTokenList(const TokenList *list) const = 0;
   virtual void OutputToken(const ANTLRTokenPtr& tokPtr) const = 0;

   // pragma processing
   virtual void HandlePragma() = 0;
>>


///////////////////////////////////////////////////////////////////////////
// Token classes
//
///////////////////////////////////////////////////////////////////////////

// Token classes are mostly used to find tokens that
// DON'T belong to the token class
#tokclass PND_NEWLINE { 
   TOK_PND TOK_PNDIF TOK_PNDIFDEF TOK_PNDIFNDEF TOK_PNDELSE TOK_PNDELIF TOK_PNDENDIF 
   TOK_PNDDEFINE TOK_PNDUNDEF TOK_PNDERROR TOK_PNDINCLUDE TOK_PNDLINE TOK_PNDPRAGMA TOK_NEWLINE 
}
#tokclass RPAREN_NEWLINE { TOK_RPAREN TOK_NEWLINE }
#tokclass LPAREN_RPAREN { TOK_LPAREN TOK_RPAREN }
#tokclass LRPAREN_DIRECTIVE_SLCOMMENT_NEWLINE { 
   TOK_LPAREN TOK_RPAREN TOK_SLCOMMENT TOK_NEWLINE
   TOK_PND TOK_PNDIF TOK_PNDIFDEF TOK_PNDIFNDEF TOK_PNDELSE TOK_PNDELIF TOK_PNDENDIF 
   TOK_PNDDEFINE TOK_PNDUNDEF TOK_PNDERROR TOK_PNDINCLUDE TOK_PNDLINE TOK_PNDPRAGMA
}
#tokclass GREATERTHAN_NEWLINE { TOK_GREATERTHAN TOK_NEWLINE }
#tokclass SLCOMMENT_NEWLINE { TOK_SLCOMMENT TOK_NEWLINE }

// Token class of all tokens that are potential macro names.
// Note that this includes normal identifiers plus C++ keywords
// The keyword "defined" cannot be redefined.
#tokclass LEGAL_MACRO_NAMES {
   TOK_ASM TOK_AUTO TOK_BOOL TOK_BREAK TOK_CASE TOK_CATCH TOK_CHAR TOK_CLASS TOK_CONST TOK_CONST_CAST
   TOK_CONTINUE TOK_DEFAULT TOK_DELETE TOK_DO TOK_DOUBLE TOK_DYNAMIC_CAST TOK_ELSE TOK_ENUM 
   TOK_EXPLICIT TOK_EXTERN TOK_BOOLFALSE TOK_FLOAT TOK_FOR TOK_FRIEND TOK_GOTO TOK_IF TOK_INLINE 
   TOK_INT TOK_LONG TOK_MUTABLE TOK_NAMESPACE TOK_NEW TOK_OPERATOR TOK_PRIVATE TOK_PROTECTED
   TOK_PUBLIC TOK_REGISTER TOK_REINTERPRET_CAST TOK_RETURN TOK_SHORT TOK_SIGNED TOK_SIZEOF
   TOK_STATIC TOK_STATIC_CAST TOK_STRUCT TOK_SWITCH TOK_TEMPLATE TOK_THIS TOK_THROW TOK_BOOLTRUE
   TOK_TRY TOK_TYPEDEF TOK_TYPEID TOK_TYPENAME TOK_UNION TOK_UNSIGNED TOK_USING TOK_VIRTUAL TOK_VOID
   TOK_VOLATILE TOK_WCHAR_T TOK_WHILE TOK_ID
   TOK_MSEXT_BASED TOK_MSEXT_DECLSPEC TOK_MSEXT_CDECL TOK_MSEXT_FASTCALL TOK_MSEXT_STDCALL
   TOK_MSEXT_INT8 TOK_MSEXT_INT16 TOK_MSEXT_INT32 TOK_MSEXT_INT64
   TOK_MSEXT_TRY TOK_MSEXT_EXCEPT TOK_MSEXT_FINALLY TOK_MSEXT_LEAVE TOK_MSEXT_INLINE
}

#tokclass MSEXT_TOKEN_CLASS {
   TOK_MSEXT_BASED TOK_MSEXT_DECLSPEC TOK_MSEXT_CDECL TOK_MSEXT_FASTCALL TOK_MSEXT_STDCALL
   TOK_MSEXT_INT8 TOK_MSEXT_INT16 TOK_MSEXT_INT32 TOK_MSEXT_INT64
   TOK_MSEXT_TRY TOK_MSEXT_EXCEPT TOK_MSEXT_FINALLY TOK_MSEXT_LEAVE TOK_MSEXT_INLINE
   TOK_MSEXT_ASM
}

///////////////////////////////////////////////////////////////////////////
// Grammar 
//
///////////////////////////////////////////////////////////////////////////

// Call this rule to process all of the input
preprocessing_file : 
   { group }
   TOK_Eof 
   ;

// Call this rule to process a single chunk of input
preprocessing_chunk > [ bool keepGoing ] : 
   <<$keepGoing = 1;>>       // init action
      group_part
   |  TOK_Eof            << $keepGoing = 0; >>
   ;

group :
   (group_part)+
   ;

group_part :
      pp_line
   |  if_line
   |  ifdef_line
   |  ifndef_line
   |  elif_line
   |  else_line
   |  endif_line
   |  control_line
   ;

// pp_line is the "generic" line; i.e. anything that is not a 
// preprocessor directive.  Note that the definition of "line"
// is augmented to include macro invocations that span lines.
pp_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList);>>
   { token_list_for_ppline [IfBlockTrue()] }
   TOK_NEWLINE
   << 
      if (IfBlockTrue()) 
      {
         ExpandAndOutputTokenList(tempTokenList); 
      }
   >>
   ;

// The constant expression part of #if or #elif must be macro-expanded
// prior to evaluation. I just collect all of the tokens up to newline,
// expand them into a new set of tokens, and then ship them off to a
// separate preprocessor-expression parser for evaluation
if_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList);>>
   tokPndif:TOK_PNDIF 
   token_list[IfBlockTrue()]
   <<
      if (IfBlockTrue())
      {
         EnterIfBlock(EvaluateExpression(), $tokPndif);
      } else {
         // Already skipping -- do not evaluate expression
         EnterIfBlock(0, $tokPndif);
      }
   >>
   TOK_NEWLINE
   ;

ifdef_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   tokPndif:TOK_PNDIFDEF 
   macroName:LEGAL_MACRO_NAMES 
   <<
      if (IfBlockTrue())
      {
         EnterIfBlock(MacroIsDefined($macroName), $tokPndif);
      } else {
         // Already skipping -- do not evaluate #ifdef
         EnterIfBlock(0, $tokPndif);
      }
   >>
   { TOK_SLCOMMENT }
   TOK_NEWLINE
   ;

ifndef_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   tokPndif:TOK_PNDIFNDEF 
   macroName:LEGAL_MACRO_NAMES 
   <<
      if (IfBlockTrue())
      {
         EnterIfBlock(!MacroIsDefined($macroName), $tokPndif);
      } else {
         // Already skipping -- do not evaluate #ifndef
         EnterIfBlock(0, $tokPndif);
      }
   >>
   { TOK_SLCOMMENT }
   TOK_NEWLINE
   ;

// See above note for if_group
elif_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   tokPndif:TOK_PNDELIF 
   token_list[EnclosingIfBlockTrue()]
   <<
      if (EnclosingIfBlockTrue())
      {
         EnterElifBlock(EvaluateExpression(), $tokPndif);
      }
   >>
   TOK_NEWLINE
   ;

else_line :
   tokPndelse:TOK_PNDELSE 
   <<
      if (EnclosingIfBlockTrue())
      {
         EnterElseBlock($tokPndelse);
      }
   >>
   { TOK_SLCOMMENT }
   TOK_NEWLINE
   ;

endif_line :
   tokPndendif:TOK_PNDENDIF 
   <<ExitIfBlock($tokPndendif);>>
   { TOK_SLCOMMENT }
   TOK_NEWLINE
   ;

// The tokens making up the include filename must be macro-expanded
// prior to inclusion. I just collect all of the tokens up to newline,
// expand them into a new set of tokens, and then concatenate the text
// of the tokens and perform some simple string-processing to extract
// the filename from the results.

control_line :
      include_line
   |  define_line
   |  undef_line
   |  line_line
   |  error_line
   |  pragma_line
   |  TOK_PND { TOK_SLCOMMENT } TOK_NEWLINE
   ;

include_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   tokPndinclude:TOK_PNDINCLUDE 
   token_list[IfBlockTrue()] TOK_NEWLINE
   <<
      if (IfBlockTrue()) 
      {
         IncludeFile($tokPndinclude);
      }
   >>
   ;

define_line :
   TOK_PNDDEFINE
   (
         function_macro_definition
      |  id_macro_definition
   )
   ;

function_macro_definition :
   // NOTE: Strip trailing '(' from funcid to get symbol
   <<;/*init-action*/>>
   <<
      ClearTokenList(tempTokenList);
      ClearTokenList(tempIdentifierList); 
   >>
   // do not need LEGAL_MACRO_NAMES because of special lexclass used here
   tokFuncid:TOK_FUNCID       
   { identifier_list[IfBlockTrue()] }
   TOK_RPAREN 
   { token_list[IfBlockTrue()] } 
   TOK_NEWLINE
   << if (IfBlockTrue()) EnterFunctionMacro($tokFuncid); >>
   ;

id_macro_definition :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   // do not need LEGAL_MACRO_NAMES because of special lexclass used here
   tokId:TOK_ID 
   { token_list[IfBlockTrue()] } 
   TOK_NEWLINE
   << if (IfBlockTrue()) EnterSimpleMacro($tokId); >>
   ;

undef_line :
   TOK_PNDUNDEF 
   macroName:LEGAL_MACRO_NAMES 
   { TOK_SLCOMMENT } 
   TOK_NEWLINE
   << if (IfBlockTrue()) UndefineMacro($macroName); >>
   ;

line_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   TOK_PNDLINE 
   token_list[false] 
   TOK_NEWLINE
   ;

error_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   TOK_PNDERROR 
   { token_list[IfBlockTrue()] } 
   TOK_NEWLINE
   << if (IfBlockTrue()) HandlePndError(); >>
   ;

pragma_line :
   <<;/*init-action*/>>
   <<ClearTokenList(tempTokenList); >>
   tokPndpragma:TOK_PNDPRAGMA 
   << AppendToTokenList(tempTokenList, $tokPndpragma); >>
   {token_list[IfBlockTrue()]} 
   << 
      if (IfBlockTrue() && GetOption(OptPragmas))
      {
         // Process the pragma
         HandlePragma();
      }
   >>
   TOK_NEWLINE
   ;


// Token lists and identifier lists are stored in class-data-members.
// Thus, a rule cannot refer to more than one identifier_list or token_list
// without copying the list data member after parsing the first list.
// This is something of an architectural wort and would not be used for
// anything more complex than the preprocessor.
// However this is fast, multithread-safe, and does not leak from exceptions 
// or guess mode longjmps.

// Note: parent rules are responsible for clearing the list data
// members before these subrules are invoked, to avoid appending to
// any previously-processed lists.

identifier_list [ bool collecting ] : 
   id:LEGAL_MACRO_NAMES
   <<if ($collecting) AppendToTokenList(tempIdentifierList, $id);>> 
   (
      TOK_COMMA 
      id2:LEGAL_MACRO_NAMES
      <<if ($collecting) AppendToTokenList(tempIdentifierList, $id2);>> 
   )*
   ;


// These rules are used to collect a "line" of tokens for a
// pp_line.  However, since macro invocations may span lines,
// this rule will keep collecting tokens across a newline if
// needed to complete a macro invocation.  We *could* just
// collect all tokens until a directive line and avoid the problem,
// but that would require storing thousands of tokens.
// There may be simpler or better approaches, but this seems
// to address the fundamental problem of resolving a combination
// of line-orientation and structure-orientation exhibited in the
// preprocessor macro-replacement semantics.

token_list_for_ppline [ bool collecting ] : 
   <<;>>    // avoid predicate hoisting
   (
         // Function-like macros may span lines but we only care if we
         // are collecting tokens

         << $collecting && IsFunctionLikeMacro(LT(1)) >>?

         macroName:LEGAL_MACRO_NAMES 
         <<AppendToTokenList(tempTokenList, $macroName);>>
      
         #pragma approx   // ambiguity OK here
         (TOK_NEWLINE | TOK_SLCOMMENT)*

         #pragma approx   // ambiguity OK here
         {
            // optional macro argument list
            parenthesized_token_list
         }
      |

         ~pndNewline:PND_NEWLINE
         <<
            if ($collecting) 
            {
               AppendToTokenList(tempTokenList, $pndNewline);
            }
         >>
   )+
   ;

parenthesized_token_list :
   lparen:TOK_LPAREN
   <<AppendToTokenList(tempTokenList, $lparen);>> 
   token_list_no_rparen 
   rparen:TOK_RPAREN
   <<AppendToTokenList(tempTokenList, $rparen);>> 
   ;

token_list_no_rparen :
   (
         parenthesized_token_list
      |  
         TOK_SLCOMMENT                                // do not add comments to list
      |  
         ~whatever:LRPAREN_DIRECTIVE_SLCOMMENT_NEWLINE
         <<AppendToTokenList(tempTokenList, $whatever);>> 
      | 
         TOK_NEWLINE                                   // do not add newlines to token list
   )*
   ;

// The token_list rule is used in cases where macro 
// invocations are not allowed to straddle newlines 
// (i.e., in directives).

token_list [ bool collecting ] : 
   (
         TOK_SLCOMMENT           // don't collect comments
      |  
         ~whatever:SLCOMMENT_NEWLINE  
         <<if ($collecting) AppendToTokenList(tempTokenList, $whatever);>> 
   )+
   ;

}
