/*
 * PUBLIC DOMAIN PCCTS-BASED C++ PREPROCESSOR GRAMMAR
 *
 * This is a companion preprocessor designed to work closely with
 * the PCCTS-based C++ grammar.  See the README for more information.
 *
 * Preprocessor Grammar:
 * Author:  
 *       John Lilley, Empathy Software     
 *       jlilley@empathy.com
 *       http://www.empathy.com
 *
 * Portions of preprocessor derived from the corresponding
 * public domain C++ grammar courtesy of the following people:
 *      Sumana Srinivasan, NeXT Inc.;               sumana_srinivasan@next.com
 *      Terence Parr, Parr Research Corporation;    parrt@parr-research.com
 *      Russell Quong, Purdue University;           quong@ecn.purdue.edu
 *
 * VERSION 1.0
 *
 * SOFTWARE RIGHTS
 * These software rights apply to the portions of this public-domain
 * C++ preprocessor grammar developed by either party listed above.
 * 
 * This file is a part of the ANTLR-based C++ grammar and is free
 * software.  We do not reserve any LEGAL rights to its use or
 * distribution, but you may NOT claim ownership or authorship of this
 * grammar or support code.  An individual or company may otherwise do
 * whatever they wish with the grammar distributed herewith including the
 * incorporation of the grammar or the output generated by ANTLR into
 * commerical software.  You may redistribute in source or binary form
 * without payment of royalties to us as long as this header remains
 * in all source distributions.
 *
 * We encourage users to develop parsers/tools using this grammar.
 * In return, we ask that credit is given to us for developing this
 * grammar.  By "credit", we mean that if you incorporate our grammar or
 * the generated code into one of your programs (commercial product,
 * research project, or otherwise) that you acknowledge this fact in the
 * documentation, research report, etc....  In addition, you should say nice
 * things about us at every opportunity.
 *
 * As long as these guidelines are kept, we expect to continue enhancing
 * this grammar.  Feel free to send us enhancements, fixes, bug reports,
 * suggestions, or general words of encouragement at parrt@parr-research.com.
 *
 * C++ Preprocessor questions and concerns should be sent to:
 *      jlilley@empathy.com
 * 
 * C++ Grammar conceived in conjunction with a project at:
 *      NeXT Computer Inc.
 *      900 Chesapeake Dr.
 *      Redwood City, CA 94555
 * 
 * Built and tested with PCCTS version 1.33
 * 
 * DISCLAIMER: we make no guarantees that this grammar works, makes sense,
 * or can be used to do anything useful.
 * [ John Lilley ] ... but if it is broken, I or others can help on
 * mercenary terms  :-)
 *
 */

#header <<
#include "FastToken.h"
#include "ATokPtr.h"
//typedef FastToken ANTLRToken;

#ifdef _MSC_VER
   // MSVC++ doesn't grok bool, so include an STL header since they define
   // bool and we'd like the same bool defined everywhere.  This works
   // for both Rogue Wave and MSVC++ STL.
   #include <new>
#endif

#include "cpre_expr_support.h"
>>

// Token definitions generated by preprocessor grammar
#tokdefs "tokens.h"

#lexclass START

class CPreExpr {
<<
   public:
      virtual void init();
   protected:
      CPreValue ConvertOctal(ANTLRTokenPtr token) const;
      CPreValue ConvertDecimal(ANTLRTokenPtr token) const;
      CPreValue ConvertHexadecimal(ANTLRTokenPtr token) const;
      CPreValue ConvertCharacter(ANTLRTokenPtr token) const;
      CPreValue ConvertFloat(ANTLRTokenPtr token) const;

>>

/////////////////////// EXPRESSION GRAMMAR ////////////////////////////

#tokclass UNARY_OPERATOR { TOK_PLUS TOK_MINUS TOK_TILDE TOK_NOT }
#tokclass MULTIPLICATIVE_OPERATOR { TOK_STAR TOK_DIVIDE TOK_MOD }
#tokclass ADDITIVE_OPERATOR { TOK_PLUS TOK_MINUS }
#tokclass SHIFT_OPERATOR { TOK_SHIFTLEFT TOK_SHIFTRIGHT }
#tokclass RELATIONAL_OPERATOR { TOK_LESSTHAN TOK_GREATERTHAN TOK_LESSTHANOREQUALTO TOK_GREATERTHANOREQUALTO }
#tokclass EQUALITY_OPERATOR { TOK_EQUAL TOK_NOTEQUAL }

// The following is an entirely separate grammar and parser, not just
// a set of subrules for the preprocessor.  When the preprocessor needs
// to evaluate a constant expression, it will invoke an instance of the
// expression parser and ask it to parse the expression.  This cannot
// be part of the preprocessor grammar, because lists of tokens must
// be macro-expanded before expressions are parsed.


// constant_expression is a conditional expression

constant_expression > [ CPreValue val ] :
   logical_or_expression          > [ $val ]
   { 
      TOK_QUESTIONMARK 
      logical_or_expression       > [ CPreValue operand1 ]
      TOK_COLON 
      logical_or_expression       > [ CPreValue operand2 ]
      <<
         bool condition = ($val.IsInteger() ? ((long)$val != 0) : ((double)$val != 0.0));
         $val = condition ? operand1 : operand2;
      >>
   }
   ;

logical_or_expression > [ CPreValue val ] : 
   logical_and_expression           > [ $val ]
   (
      TOK_OR 
      logical_and_expression        > [ CPreValue operand ]
      <<
         // Logical TOK_OR must use integers
         $val = CPreValue( (long)((long)$val || (long)operand) );
      >>
   )*
   ;

logical_and_expression > [ CPreValue val ] :
   inclusive_or_expression           > [ $val ]
   (
      TOK_AND 
      inclusive_or_expression        > [ CPreValue operand ]
      <<
         // Logical TOK_AND must use integers
         $val = CPreValue( (long)((long)$val && (long)operand) );
      >>
   )*
   ;

inclusive_or_expression > [ CPreValue val ] :
   exclusive_or_expression           > [ $val ]
   (
      TOK_BITWISEOR 
      exclusive_or_expression        > [ CPreValue operand ]
      <<
         // Bitwise TOK_OR must use integers -- invalid for doubles, but try anyway
         $val = CPreValue( (long)$val | (long)operand );
      >>
   )*
   ;

exclusive_or_expression > [ CPreValue val ] :
   and_expression           > [ $val ]
   (
      TOK_BITWISEXOR 
      and_expression        > [ CPreValue operand ]
      <<
         // Bitwise XOR must use integers -- invalid for doubles, but try anyway
         $val = CPreValue( (long)$val ^ (long)operand );
      >>
   )*
   ;

and_expression > [ CPreValue val ] :
   equality_expression           > [ $val ]
   (
      TOK_AMPERSAND 
      equality_expression        > [ CPreValue operand ]
      <<
         // Bitwise TOK_AND must use integers -- invalid for doubles, but try anyway
         $val = CPreValue( (long)$val & (long)operand );
      >>
   )*
   ;

equality_expression > [ CPreValue val ] :
   relational_expression           > [ $val ]
   (
      eo:EQUALITY_OPERATOR
      relational_expression        > [ CPreValue operand ]
      <<
         switch ($eo->getType())
         {
            case TOK_NOTEQUAL:
               // Integers promoted to double can be compared with the same result
               $val = CPreValue( (long)((double)$val != (double)operand) );
               break;
            case TOK_EQUAL:
               // Integers promoted to double can be compared with the same result
               $val = CPreValue( (long)((double)$val == (double)operand) );
               break;
         }
      >>
   )*
   ;

relational_expression > [ CPreValue val ] :
   shift_expression              > [ $val ]
   (  
      ro:RELATIONAL_OPERATOR
      shift_expression           > [ CPreValue operand ]
      <<
         switch ($ro->getType())
         {
            case TOK_LESSTHAN:
               $val = CPreValue( (long)((double)$val < (double)operand) );
               break;
            case TOK_GREATERTHAN:
               $val = CPreValue( (long)((double)$val > (double)operand) );
               break;
            case TOK_LESSTHANOREQUALTO:
               $val = CPreValue( (long)((double)$val <= (double)operand) );
               break;
            case TOK_GREATERTHANOREQUALTO:
               $val = CPreValue( (long)((double)$val >= (double)operand) );
               break;
         }
      >>
   )*
   ;

shift_expression > [ CPreValue val ] :
   additive_expression            > [ $val ]
   (
      so:SHIFT_OPERATOR
      additive_expression         > [ CPreValue operand ]
      <<
         // Slow, but shift in #if expression is rare
         long shiftamount = (long)operand;
         long value = (long)$val;
         if ($so->getType() == TOK_SHIFTRIGHT)
         {
            shiftamount = -shiftamount;
         }
         if (shiftamount < -64) shiftamount = -64;
         if (shiftamount > 64) shiftamount = 64;
         while (shiftamount > 0)
         {
            value *= 2;
            shiftamount--;
         }
         while (shiftamount < 0)
         {
            value /= 2;
            shiftamount++;
         }
         $val = CPreValue(value);
      >>
   )*
   ;

additive_expression > [ CPreValue val ] :
   multiplicative_expression      > [ $val ]
   (
      ao:ADDITIVE_OPERATOR
      multiplicative_expression   > [ CPreValue operand ]
      <<
         switch ($ao->getType())
         {
            case TOK_PLUS:
               if (!$val.IsInteger() || !operand.IsInteger())
               {
                  $val = CPreValue((double)$val + (double)operand);
               } else {
                  $val = CPreValue((long)$val + (long)operand);
               }
               break;
            case TOK_MINUS:
               if (!$val.IsInteger() || !operand.IsInteger())
               {
                  $val = CPreValue((double)$val - (double)operand);
               } else {
                  $val = CPreValue((long)$val - (long)operand);
               }
               break;
         }
      >>
   )*
   ;

multiplicative_expression > [ CPreValue val ] :
    unary_expression          > [ $val ]
	 (
      mo:MULTIPLICATIVE_OPERATOR
      unary_expression        > [ CPreValue operand ]
      <<
         switch ($mo->getType())
         {
            case TOK_STAR:
               if (!$val.IsInteger() || !operand.IsInteger())
               {
                  return CPreValue((double)$val * (double)operand);
               } else {
                  return CPreValue((long)$val * (long)operand);
               }
               break;
            case TOK_DIVIDE:
               if (!$val.IsInteger() || !operand.IsInteger())
               {
                  return CPreValue((double)$val + ((double)operand == 0.0 ? 1.0 : (double)operand));
               } else {
                  long operandval = (long)operand;
                  if (operandval==0) operandval = 1;
                  return CPreValue((long)$val / operandval);
               }
               break;
            case TOK_MOD:
               {
                  // Mod only makes sense for integers
                  long operandval = (long)operand;
                  if (operandval==0) operandval = 1;
                  return CPreValue((long)$val % operandval);
               }
               break;
         }
      >>
    )*
   ;

unary_expression  > [ CPreValue val ] :
      primary_expression      > [ $val ]
	|  
      uo:UNARY_OPERATOR
      unary_expression > [ CPreValue operand ]
      <<
         switch ($uo->getType())
         {
            case TOK_PLUS:
               $val = operand;
               break;
            case TOK_MINUS:
               if (operand.IsInteger())
               {
                  $val = CPreValue(-(long)operand);
               } else {
                  $val = CPreValue(-(double)operand);
               }
               break;
            case TOK_TILDE:
               // ~ cannot apply to double value, but process doubles as longs anyway
               $val = CPreValue(~(long)operand);
               break;
            case TOK_NOT:
               $val = CPreValue((long)!(long)operand);
               break;
         }
      >>
   ;

primary_expression > [ CPreValue val ] :
      constant                > [ $val ]
   |  TOK_LPAREN 
      constant_expression     > [ $val ]
      TOK_RPAREN
   ;

constant > [ CPreValue val ] :
      oct:TOK_OCTALINT          << $val = ConvertOctal($oct); >>
   |  dec:TOK_DECIMALINT        << $val = ConvertDecimal($dec); >>
   |  hex:TOK_HEXADECIMALINT    << $val = ConvertHexadecimal($hex); >>
   |  chr:TOK_CHARACTER         << $val = ConvertCharacter($chr); >>
   |  fl1:TOK_FLOATONE          << $val = ConvertFloat($fl1); >>
   |  fl2:TOK_FLOATTWO          << $val = ConvertFloat($fl2); >>
   ;


// "defined id" is handled outside of expression parser, since it is
// replaced before macroe expansion of the expression tokens.
//defined_args :
//      TOK_LPAREN TOK_ID TOK_RPAREN
//   |  TOK_ID
//   ;

// Note: There is no cast expression.  It is forbidden by the standard,
// not supported by MSVC++, and the expression:
//    #if (long)(int)1 
// crashes borland C++ 5.0
//cast_expression :
//      TOK_LPAREN builtin_type TOK_RPAREN cast_expression
//   |  unary_expression
//   ;
//
//builtin_type :
//   (
//         TOK_CHAR
//      |  TOK_SHORT
//      |  TOK_INT
//      |  TOK_LONG
//      |  TOK_FLOAT
//      |  TOK_DOUBLE
//      |  TOK_SIGNED
//      |  TOK_UNSIGNED
//   )+
//   ;

}

