// cpre_expr_support.cpp:  Support class implementations for the C preprocesor
//    expression evaluator.

#include "tokens.h"
#include "CPreExpr.h"
#include "cpre_expr_support.h"
#include "assert.h"
#include "math.h"

///////////////////////////////////////////////////////////////////////
//                         Construction                              //
///////////////////////////////////////////////////////////////////////
void CPreExpr::init()
{
   ANTLRParser::init();
}

///////////////////////////////////////////////////////////////////////
//                    Conversion methods                             //
///////////////////////////////////////////////////////////////////////

CPreValue CPreExpr::ConvertOctal(ANTLRTokenPtr token) const
{
   long val = 0;
   for (const char *p = token->getText(); *p != 0; p++)
   {
      if (*p >= '0' && *p <= '7')
      {
         val = val * 8 + *p - '0';
      }
   }
   return CPreValue(val);
}

CPreValue CPreExpr::ConvertDecimal(ANTLRTokenPtr token) const
{
   long val = 0;
   for (const char *p = token->getText(); *p != 0; p++)
   {
      if (*p >= '0' && *p <= '9')
      {
         val = val * 10 + *p - '0';
      }
   }
   return CPreValue(val);
}


CPreValue CPreExpr::ConvertHexadecimal(ANTLRTokenPtr token) const
{
   long val = 0;
   for (const char *p = token->getText(); *p != 0; p++)
   {
      if (*p >= '0' && *p <= '9')
      {
         val = val * 16 + *p - '0';
      } else if (*p >= 'a' && *p <= 'f')
      {
         val = val * 16 + *p - 'a' + 10;
      } else if (*p >= 'A' && *p <= 'F')
      {
         val = val * 16 + *p - 'A' + 10;
      }
   }
   return CPreValue(val);
}


CPreValue CPreExpr::ConvertCharacter(ANTLRTokenPtr token) const
{
   return CPreValue((long)token->getText()[1]);
}


CPreValue CPreExpr::ConvertFloat(ANTLRTokenPtr token) const
{
   return CPreValue(atof(token->getText()));
}


