// cpre_ifblock.cpp: handling if nested #if/else/endif block structures

#include "CPreParserImp.h"
#include "cpre_ifblock.h"

///////////////////////////////////////////////////////////////////////
//                    If block control                               //
///////////////////////////////////////////////////////////////////////
void CPreParserImp::EnterIfBlock(bool isTrue, const ANTLRTokenPtr& )
{
   bool enclosingTrue = IfBlockTrue();
   // If enclosing block is false, then this block is also false
   ifBlockStack->push(IfBlock(isTrue && enclosingTrue, enclosingTrue));
}

void CPreParserImp::EnterElifBlock(bool isTrue, const ANTLRTokenPtr& token)
{   
   if (!InsideIfPart())
   {
      HandleError("#elif without matching #if", token);
      return;
   }
   if (EnclosingIfBlockTrue())
   {
      if (IfBlockTrue())
      {
         // entered a (false) #elif block from a true block
         ifBlockStack->top().change(false);
      } else if (isTrue && !ifBlockStack->top().SomePartWasTrue())
      {
         // Entered true #elif block and no previous block was true
         ifBlockStack->top().change(isTrue);
      }
   }
}

void CPreParserImp::EnterElseBlock(const ANTLRTokenPtr& token)
{
   if (!InsideIfPart())
   {
      HandleError("#else without matching #if", token);
      return;
   }
   if (EnclosingIfBlockTrue())
   {
      if (!ifBlockStack->top().SomePartWasTrue())
      {
         // Entered (true) #else block and no previous block was true
         ifBlockStack->top().change(true);
      } else if (IfBlockTrue())
      {
         // Entered (false) else block from true block
         ifBlockStack->top().change(false);
      }

      // Set else flag
      ifBlockStack->top().SetElse();
   }
}

void CPreParserImp::ExitIfBlock(const ANTLRTokenPtr& token)
{
   if (ifBlockStack->size() == 0)
   {
      HandleError("#endif without matching #if", token);
      return;
   }
   ifBlockStack->pop();
}

bool CPreParserImp::IfBlockTrue() const
{ 
   return ifBlockStack->size() == 0 || ifBlockStack->top().IsTrue(); 
}

bool CPreParserImp::EnclosingIfBlockTrue() const
{
   return ifBlockStack->size() == 0 || ifBlockStack->top().EnclosingTrue();
}

bool CPreParserImp::InsideElsePart() const
{
   return ifBlockStack->size() != 0 && ifBlockStack->top().InElse();
}

bool CPreParserImp::InsideIfPart() const
{
   return ifBlockStack->size() != 0 && !ifBlockStack->top().InElse();
}

