// cpre_ifblock.h:  Class to maintain state of nested #if/#else/#endif blocks.

#ifndef INCL_CPRE_IFBLOCK_H
#define INCL_CPRE_IFBLOCK_H

#include <stack>
#include <deque>

///////////////////////////////////////////////////////////////////////
//                       if block support                            //
///////////////////////////////////////////////////////////////////////
class IfBlock
{
public:
   IfBlock(bool isTrue_, bool enclosingTrue_) : 
      isTrue(isTrue_), 
      somePartWasTrue(isTrue_),
      enclosingTrue(enclosingTrue_),
      inElse(0)
   {}

   // These are only here because MSVC++ is a sick compiler, and insists on
   // performing semantic checks on non-instantiated container template 
   // methods.  These are never used.
   IfBlock() { assert(0); }
   bool operator == (const IfBlock& rhs) const { assert(0); return 0; }
   bool operator < (const IfBlock& rhs) const { assert(0); return 0; }

   void change(bool isTrue_) { isTrue = isTrue_; if (isTrue_) somePartWasTrue = 1; }
   bool IsTrue() const { return isTrue; }
   bool SomePartWasTrue() const { return somePartWasTrue; }
   bool EnclosingTrue() const { return enclosingTrue; }
   bool InElse() const { return inElse; }
   void SetElse() { inElse = 1; }
private:
   bool isTrue;             // Current block is true
   bool somePartWasTrue;    // One of the preceeding or current #if/#elif was true
   bool enclosingTrue;      // Enclosing #if block is true
   bool inElse;             // Inside the #else part
};

class IfBlockStack : public stack< IfBlock, deque<IfBlock> >
{
};

#endif    //INCL_CPRE_IFBLOCK_H
