// cpre_include.h:  Include processing support

#ifndef INCL_CPRE_INCLUDE_H
#define INCL_CPRE_INCLUDE_H

#include <stack>
#include <list>
#include <assert.h>
#include <stdio.h>

class DLGInputStream ;
class DLGLexer;
class ANTLRAbstractToken;

///////////////////////////////////////////////////////////////////////
//              record containing include file information           //
///////////////////////////////////////////////////////////////////////

// IncludeRecord: This struct contains the FILE, input stream, and
// lexer pointers used during file inclusion.  When the EOF is hit
// on an included file, the FILE, stream, and lexer will be deleted.
struct IncludeRecord {
   IncludeRecord(
      FILE *fp_, 
      DLGInputStream *stream_, 
      DLGLexer *scan_, 
      ANTLRAbstractToken* tok_
   ) :
      fp(fp_), stream(stream_), scan(scan_), tok(tok_)
   {}

   // These are only here because MSVC++ is a sick compiler, and insists on
   // performing semantic checks on non-instantiated container template 
   // methods.  These are never used.
   IncludeRecord() { assert(0); }
   bool operator == (const IncludeRecord& rhs) const { assert(0); return 0; }
   bool operator < (const IncludeRecord& rhs) const { assert(0); return 0; }

   FILE *fp;
   DLGInputStream  *stream;
   DLGLexer *scan;
   JLStr filename;
   ANTLRAbstractToken *tok;       // token passed to lexer to create new tokens
};

// Can omit last two template parms when MSVC++ supports default template args
typedef stack<IncludeRecord, deque<IncludeRecord> > IncludeRecordStack;

///////////////////////////////////////////////////////////////////////
//              Include Path management
///////////////////////////////////////////////////////////////////////

typedef list<JLStr> IncludePath;

#endif
