// cpre_init.cpp: construction and initialization methods

#include "CPreParserImp.h"
#include "cpre_ifblock.h"
#include "cpre_include.h"
#include "cpre_symtable.h"
#include "TokenList.h"
#include <map>
#include <set>

///////////////////////////////////////////////////////////////////////
//                         Construction                              //
///////////////////////////////////////////////////////////////////////
CPreParserImp::CPreParserImp(
   ANTLRTokenStream *input_
) : 
   CPreParser(&inputStack),
   inputStack(input_),
   simpleSymtable(0),
   functionSymtable(0),
   ifBlockStack(0),
   includeStack(0),
   includePath(0),
   macrosBeingExpanded(0),
   options(0)
{
   // Let the input stack communicate end-of-include-file to the parser
   inputStack.SetParser(this);
}

void CPreParserImp::init()
{
   CPreParser::init();

   simpleSymtable = new MacroSymbolTable<MacroEntrySimpleRef>;
   functionSymtable = new MacroSymbolTable<MacroEntryFunctionRef>;
   ifBlockStack = new IfBlockStack;
   includeStack = new IncludeRecordStack;
   includePath = new IncludePath;
   macrosBeingExpanded = new CPreMacroSet;
   options = new OptionSet;

   // Initialize default options
   SetOption(OptMSExtensions, false);
   SetOption(OptBool, true);
   SetOption(OptWCharT, true);
   SetOption(OptPragmas, true);
   SetOption(OptExpandPragmas, true);
   SetOption(OptTrackInclude, false);
   SetOption(OptTypename, true);

   outputFile = stdout;
}

void CPreParser::init()
{
   ANTLRParser::init();
   tempTokenList = new TokenList;
   tempIdentifierList = new TokenList;
}

///////////////////////////////////////////////////////////////////////
//                         Destruction                               //
///////////////////////////////////////////////////////////////////////
CPreParserImp::~CPreParserImp()
{
   delete macrosBeingExpanded; 
   macrosBeingExpanded = 0;
   delete includePath; 
   includePath = 0;
   delete includeStack; 
   includeStack = 0;
   delete ifBlockStack; 
   ifBlockStack =0;
   delete functionSymtable; 
   functionSymtable = 0;
   delete simpleSymtable; 
   simpleSymtable = 0;
   delete options;
   options = 0;
}

CPreParser::~CPreParser()
{
   delete tempIdentifierList;
   tempIdentifierList = 0;
   delete tempTokenList;
   tempTokenList = 0;
}

