// cpre_output.cpp: default output of tokens and token lists

#include "CPreParserImp.h"
#include "tokenList.h"

///////////////////////////////////////////////////////////////////////
// OutputTokenBase: non-virtual method called to output a token.
// It first performs any mappings needed to support specific compiler
// options, and then calls OutputToken() to perform the actual output.
//
// Inputs:
//    ANTLRTokenPtr&    token    Token to output
// Outputs:
//    ANTLRTokenPtr&    token    Token modified as necessary to
//                               support parser options.
// Return: none
///////////////////////////////////////////////////////////////////////
void 
CPreParserImp::OutputTokenBase(
   ANTLRTokenPtr& token
) const
{
   switch (token->getType())
   {
      case TOK_BOOL:
      case TOK_BOOLFALSE:
      case TOK_BOOLTRUE:
         if (!GetOption(OptBool))
         {
            token->setType(TOK_ID);
         }
         break;
      case TOK_WCHAR_T:
         if (!GetOption(OptWCharT))
         {
            token->setType(TOK_ID);
         }
         break;
      case TOK_TYPENAME:
         if (!GetOption(OptTypename))
         {
            token->setType(TOK_ID);
         }
         break;
      default:
         if (set_el(token->getType(), MSEXT_TOKEN_CLASS_set))
         {
            if (!GetOption(OptMSExtensions))
            {
               token->setType(TOK_ID);
            }
         }
         break;
   }

   OutputToken(token);
}

///////////////////////////////////////////////////////////////////////
// OuptutToken: virtual method to perform actual token output.  Can be 
//              overridden to output to some other destination.
//
// Inputs:
//    ANTLRTokenPtr&    token    Token to output
// Outputs: none
// Return: none
///////////////////////////////////////////////////////////////////////
void 
CPreParserImp::OutputToken(
   const ANTLRTokenPtr& token
) const
{
   fprintf(outputFile, "%s ", token->getText());
}

///////////////////////////////////////////////////////////////////////
// OuptutTokenList:  Output a list of tokens.  Performs some minimal
//       processing to handle TOK_SLCOMMENT tokens generated from ##
//       operations, and calls OutputTokenBase to output each token.
//
// Inputs:
//    const TokenList*     tokenList      List of tokens to output
// Outputs: none
// Return: none
///////////////////////////////////////////////////////////////////////
void 
CPreParserImp::OutputTokenList(
   const TokenList *tokenList
) const
{
   for(
      TokenList::const_iterator iter(tokenList->begin());
      iter != tokenList->end();
      iter++
   )
   {
      ANTLRTokenPtr token = *iter;
      if (token->getType() == TOK_SLCOMMENT)
      {
         // Skip rest of line when TOK_SLCOMMENT is encountered.
         // This is to support obnoxious definition in MS-Windows "richedit.h":
         //     define  _WPAD   /##/
         break;
      }
      OutputTokenBase(token);
   }
}

