// DebugOutput.h: header file for output debug window class

#ifndef INCL_DEBUGOUTPUT_H
#define INCL_DEBUGOUTPUT_H

#include <afx.h>

// The class that really does the work
class CDebugOutputImp;


// Wrapper object over CDebugOutputImp
class CDebugOutput
{
public:
   // If outputFilename is non-null, then debug output also goes to the file.
   CDebugOutput(
      const char* outputFilename = NULL,
      int indentSpacing = 3
   );
   ~CDebugOutput();

   // These methods are forwarded to implementation class
   void Indent();
   void Unindent();
   void Output(const CString& str);
   void Output(const char* str);

private:
   // There is no reference-counting.
   // One CDebugOutputImp per CDebugOutput.
   CDebugOutputImp* imp;
};


// Object that automatically tracks entry/exit of method if
// created as a local variable at the top of a method body.
class CDebugTracker
{
public:
   CDebugTracker(const CString& methodName_);
   ~CDebugTracker();

   // Access to the singleton debug output object used by the trackers.
   CDebugOutput& GetDebugOutput() { return output; }
   void Output(const CString& str) { output.Output(str); }

private:
   CString methodName;
   // Singleton debug output class used by all CDebugTracker objects
   static CDebugOutput output;
};


// Macro to make CDebugTracker creation depend on _DEBUG flag
#ifdef _DEBUG
#define DEBUG_TRACKER(method) CDebugTracker tracker(method);
#else
#define DEBUG_TRACKER(method)
#endif


#endif
