////////////////////////////////////////////////////////////////////////////////
// Freelist memory allocator

#ifndef INCL_FREELIST_H
#define INCL_FREELIST_H

#include "assert.h"
#include "STLIncomplete.h"

using namespace std;

/////////////////////////////////////////////////////////////////////
// JLFreeList:  Fast fixed-size memory allocator
// Note: All allocated blocks are freed when freelist is deleted!
/////////////////////////////////////////////////////////////////////
class JLFreeList
{
public:
	JLFreeList(size_t eltSize_, int blockSize_ = 100);
	~JLFreeList();
	
	void* New();

	void Delete(void* ptr)
   {
      if (ptr != 0)
      {
   	   ((JLFreeElt*)ptr)->next = freeList;
	      freeList = (JLFreeElt*)ptr;
      }
   }


private:
	// Do not allow copy / assignment
	JLFreeList(JLFreeList& ) { assert(0); }
	JLFreeList& operator=(const JLFreeList&) { assert(0); return *this;}
	
	class JLFreeElt
	{
		friend class JLFreeList;
		JLFreeElt* next;
	};
	void AddBlock();
	size_t eltSize;			// size of each element
	int blockSize;          // Number of elements in each new block
	JLFreeElt* freeList;    // threaded list of free objects
	typedef list<char*, allocator<char*> > list_of_char_ptr;
	list_of_char_ptr *blocks;
};

typedef JLFreeList* JLFreeListPtr;

///////////////////////////////////////////////////////////////////////
// JLMemPool: Fast variable-size memory allocator
// Quantizes allocations to nearest 4 bytes.
// Note: All allocations <= maxSize are freed when this is destroyed!
///////////////////////////////////////////////////////////////////////
class JLMemPool
{
public:
	JLMemPool(size_t maxSize, int blockSize_ = 50);
	~JLMemPool();

	void* New(size_t size);
	void Delete(void* ptr, size_t size);

private:
   enum { ChunkSize = 4, MaxFreeLists = 50 };
   int nbrFreeLists;
   JLFreeList** freeLists;    // ptr to array of pointers to JLFreeList
};

#endif
