// FastToken.cpp: Methods for fast token implementation

#include "tokens.h"
#include "FastToken.h"


/////////////////////////////////////////////////////////////////////
//                   Variable (large) text token                   //
/////////////////////////////////////////////////////////////////////

void FastTokenVar::setText(const ANTLRChar *s)
{
   if ((int)strlen(s) < bufSize)
   {
      strcpy(text, s);
   } else {
      delete [] text;
      InitBuffer(s);
   }
}

/////////////////////////////////////////////////////////////////////
//                Token memory pools                               //
/////////////////////////////////////////////////////////////////////
JLFreeList FastTokenFixed<4>::freeList(sizeof(FastTokenFixed<4>), FreeListBlockSize);
JLFreeList FastTokenFixed<8>::freeList(sizeof(FastTokenFixed<8>), FreeListBlockSize);
JLFreeList FastTokenFixed<12>::freeList(sizeof(FastTokenFixed<12>), FreeListBlockSize);
JLFreeList FastTokenFixed<16>::freeList(sizeof(FastTokenFixed<16>), FreeListBlockSize);
JLFreeList FastTokenFixed<20>::freeList(sizeof(FastTokenFixed<20>), FreeListBlockSize);
JLFreeList FastTokenFixed<24>::freeList(sizeof(FastTokenFixed<24>), FreeListBlockSize);
JLFreeList FastTokenFixed<28>::freeList(sizeof(FastTokenFixed<28>), FreeListBlockSize);
JLFreeList FastTokenFixed<32>::freeList(sizeof(FastTokenFixed<32>), FreeListBlockSize);
JLFreeList FastTokenFixed<36>::freeList(sizeof(FastTokenFixed<36>), FreeListBlockSize);
JLFreeList FastTokenVar::freeList(sizeof(FastTokenVar), FreeListBlockSize);
//JLFreeList FastTokenTmp::freeList(sizeof(FastTokenTmp), FreeListBlockSize);

/////////////////////////////////////////////////////////////////////
//                     Static token-making function                //
/////////////////////////////////////////////////////////////////////
ANTLRAbstractToken *
FastToken::newToken(
   ANTLRTokenType tt, 
   const char *txt, 
   int line
)
{
   switch (strlen(txt))
   {
   case 0:
   case 1:
   case 2:
   case 3:
      return new FastTokenFixed<4>(tt, txt, line);
   case 4:
   case 5:
   case 6:
   case 7:
      return new FastTokenFixed<8>(tt, txt, line);
   case 8:
   case 9:
   case 10:
   case 11:
      return new FastTokenFixed<12>(tt, txt, line);
   case 12:
   case 13:
   case 14:
   case 15:
      return new FastTokenFixed<16>(tt, txt, line);
   case 16:
   case 17:
   case 18:
   case 19:
      return new FastTokenFixed<20>(tt, txt, line);
   case 20:
   case 21:
   case 22:
   case 23:
      return new FastTokenFixed<24>(tt, txt, line);
   case 24:
   case 25:
   case 26:
   case 27:
      return new FastTokenFixed<28>(tt, txt, line);
   case 28:
   case 29:
   case 30:
   case 31:
      return new FastTokenFixed<32>(tt, txt, line);
   case 32:
   case 33:
   case 34:
   case 35:
      return new FastTokenFixed<36>(tt, txt, line);
   default:
      return new FastTokenVar(tt, txt, line);
   }
}
