#include "JLFreelist.h"
#include <list>

///////////////////////////////////////////////////////////////////////////////
//                     JLFreeList                                            //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
JLFreeList::JLFreeList(
	size_t eltSize_, 
	int blockSize_
) :
	eltSize(sizeof(JLFreeElt) > eltSize_ ? sizeof(JLFreeElt) : eltSize_),
	blockSize(blockSize_),
	freeList(0)
{
	// make sure it's somewhat reasonable
	assert((long)blockSize * (long)eltSize < 2000000L);
   blocks = new list_of_char_ptr;
}

///////////////////////////////////////////////////////////////////////////////
JLFreeList::~JLFreeList()
{
	// Note: we might want to add some bookkeeping to this, to
	// report unfreed elements
	for (
		list_of_char_ptr::iterator iter(blocks->begin());
		iter != blocks->end();
		iter++
	)
	{
		delete [] *iter;
	}
   delete blocks;
}
	
///////////////////////////////////////////////////////////////////////////////
void* 
JLFreeList::New()
{
	if (freeList == 0)
	{
		AddBlock();
	}
	void* ptr = freeList;
	freeList = freeList->next;
	return ptr;
}

///////////////////////////////////////////////////////////////////////////////
void
JLFreeList::AddBlock()
{
	char* newBlockPtr = new char[blockSize * eltSize];
	assert(newBlockPtr != 0);
	char* blockEnd = newBlockPtr + blockSize * eltSize;
	blocks->push_back(newBlockPtr);
	
	// Add the elements to the free list
	for (; newBlockPtr < blockEnd; newBlockPtr += eltSize)
	{
		((JLFreeElt*)(void*)newBlockPtr)->next = freeList;
		freeList = (JLFreeElt*)(void*)newBlockPtr;
	}
}

///////////////////////////////////////////////////////////////////////////////
//                     JLMemPool                                             //
///////////////////////////////////////////////////////////////////////////////

JLMemPool::JLMemPool(size_t maxSize, int blockSize_)
{
   nbrFreeLists = (int)maxSize / ChunkSize + 1;
   assert(nbrFreeLists > 0 && nbrFreeLists <= MaxFreeLists);
   freeLists = new JLFreeListPtr[nbrFreeLists];
   for (int i = 0; i < nbrFreeLists; i++)
   {
      freeLists[i] = new JLFreeList( (i+1)*ChunkSize, blockSize_);
   }
}

JLMemPool::~JLMemPool()
{
   for (int i = 0; i < nbrFreeLists; i++)
   {
      delete freeLists[i];
   }
   delete [] freeLists;
}

void* 
JLMemPool::New(size_t size)
{
   int idx = (int)size / ChunkSize;
   if (idx < nbrFreeLists)
   {
      return freeLists[idx]->New();
   } else {
      return new char[size];
   }
}

void 
JLMemPool::Delete(void* ptr, size_t size)
{
   if (ptr != 0)
   {
      int idx = (int)size / ChunkSize;
      if (idx < nbrFreeLists)
      {
         freeLists[idx]->Delete(ptr);
      } else {
         delete [] (char*)ptr;
      }
   }
}

