#
# package.ksh: package this directory into a gzip'd tar file,
# excluding rcs files and compilation products.
#
# usage: package.ksh [ -docs ] [ -private ] [ -date <yymmdd> ]
# -docs: include .doc files
# -private: include private tests (those that are property of others)
#

USAGE='usage: package.sh [ -docs ] [ -private ] [ -date <yymmdd> ] <outfile>'

rm -f tmp.tar tmp.tar.gz /tmp/newer.tmp /tmp/file_list.tmp

DOCSPATTERN='.*\.doc$'
NEWER=
OUTFILE=
PRIVATE_PATTERN=private_tests
while [ $# -ge 1 ]
do
	if [ $1 = "-docs" ]
	then
		DOCSPATTERN=this_should_match_nothing_at_all
		shift
	elif [ $1 = "-date" ]
	then
		shift
		if [ $# -lt 1 ]
		then
			echo $USAGE
			exit 1
		fi
		NEWER="-newer /tmp/newer.tmp"
		touch -t ${1}0000 /tmp/newer.tmp 
		shift
	elif [ $1 = "-private" ]
	then
		shift
		PRIVATE_PATTERN=DONT_MATCH_ANYTHING_AT_ALL
	elif [ $1 = "-help" ]
	then
		echo $USAGE
		exit 1
	elif [ $# = 1 ]
	then
		# must be the output file name
		OUTFILE=$1
		shift
	else
		echo $USAGE
		exit 1
	fi
done

find . -type f $NEWER -print | 
	egrep -v "$DOCSPATTERN" | 
	egrep -v "$PRIVATE_PATTERN" |
	egrep -v '.*\.exe$|.*\.obj$|.*\.lib$|.*\.ilk$|.*\.pch$|.*\.pdb$|.*\.clw$|.*\.ncb$|.*\.pj$|.*\.use$|.*\.bsc$|.*\.sbr$|.*\.idb$|.*\.aps$|.*\.res$' > /tmp/file_list.tmp

tar -cvf tmp.tar nul
cat /tmp/file_list.tmp | xargs tar -rvf tmp.tar
gzip tmp.tar

if [ $OUTFILE ]
then
	mv tmp.tar.gz $OUTFILE
fi
rm -f /tmp/newer.tmp /tmp/file_list.tmp
