/* ANTLRParser.C
 *
 * SOFTWARE RIGHTS
 *
 * We reserve no LEGAL rights to the Purdue Compiler Construction Tool
 * Set (PCCTS) -- PCCTS is in the public domain.  An individual or
 * company may do whatever they wish with source code distributed with
 * PCCTS or the code generated by PCCTS, including the incorporation of
 * PCCTS, or its output, into commerical software.
 * 
 * We encourage users to develop software with PCCTS.  However, we do ask
 * that credit is given to us for developing PCCTS.  By "credit",
 * we mean that if you incorporate our source code into one of your
 * programs (commercial product, research project, or otherwise) that you
 * acknowledge this fact somewhere in the documentation, research report,
 * etc...  If you like PCCTS and have developed a nice tool with the
 * output, please mention that you developed it using PCCTS.  In
 * addition, we ask that this header remain intact in our source code.
 * As long as these guidelines are kept, we expect to continue enhancing
 * this system and expect to make other tools available as they are
 * completed.
 *
 * ANTLR 1.33
 * Terence Parr
 * Parr Research Corporation
 * with Purdue University and AHPCRC, University of Minnesota
 * 1989-1995
 */

//#include <iostream>
#include <assert.h>

// Get the real definition of the token type
#include "tokens.h"
#include "AToken.h"
#include "ATokenBuffer.h"
#include "AParser.h"

/* maximum of 32 bits/unsigned int and must be 8 bits/byte;
 * we only use 8 bits of it.
 */
SetWordType ANTLRParser::bitmask[sizeof(SetWordType)*8] = {
	0x00000001, 0x00000002, 0x00000004, 0x00000008,
	0x00000010, 0x00000020, 0x00000040, 0x00000080
};


////////////////////////////////////////////////////////////////
// Level for trace output
int ANTLRParser::Tracer::level = 0;

////////////////////////////////////////////////////////////////
ANTLRParser::
~ANTLRParser()
{
   // delete temp token set
   delete [] tempTokenSet;
}

////////////////////////////////////////////////////////////////
ANTLRParser::
ANTLRParser(
   ANTLRTokenBuffer *inputTokens_,
	int LLk_,
	int,                // use_inf_look not used anymore
	int,                // dlook not used anymore
	int tokenSetSize_   // size of all token sets
) :
   trace_on(0),
	inputTokens(inputTokens_),
   LLk(LLk_),
	tokenSetSize(tokenSetSize_)
{
   assert(LLk < MAX_LLK);

   // allocate temp token set
   tempTokenSet = new SetWordType[tokenSetSize];

	guessing = 0;
	token_tbl = NULL;
	eofToken = (ANTLRTokenType)1;
}

////////////////////////////////////////////////////////////////
void ANTLRParser::init()
{
   prime_lookahead();
}

////////////////////////////////////////////////////////////////
void 
ANTLRParser::saveState(
   ANTLRParserState *buf
)
{
	buf->guessing = guessing;
}

////////////////////////////////////////////////////////////////
void 
ANTLRParser::restoreState(
   ANTLRParserState *buf
)
{
	guessing = buf->guessing;
   prime_lookahead();
}


////////////////////////////////////////////////////////////////
int ANTLRParser::MarkInputTokens() { return inputTokens->mark(); }
void ANTLRParser::RewindInputTokens(int marker) { inputTokens->rewind(marker); }
void ANTLRParser::UnmarkInputTokens(int marker) { inputTokens->unmark(marker); }

////////////////////////////////////////////////////////////////
// Get the next symbol from the input stream; put it into lookahead buffer;
// fill token_type[] fast reference cache also.  NLA is the next place where
// a lookahead ANTLRAbstractToken should go.
void 
ANTLRParser::consume()
{
   inputTokens->consume();
   prime_lookahead();
}

////////////////////////////////////////////////////////////////
ANTLRTokenPtr 
ANTLRParser::LT(int i)
{
	return inputTokens->LT(i);
}

////////////////////////////////////////////////////////////////
// fill the lookahead buffer up with k symbols
void
ANTLRParser::prime_lookahead()
{
	for (int i=1; i<=LLk; i++) 
   {
      token_type[i-1] = LT(i)->getType();
   }
}

////////////////////////////////////////////////////////////////
// check to see if the current input symbol matches '_t'.
void
ANTLRParser::_match(
   ANTLRTokenType _t
)
{
	if ( LA(1)!=_t ) {
      // throw an exception
      fail_syntax(_t);
   }
}

////////////////////////////////////////////////////////////////
// check to see if the current input symbol matches any token in a set.
void
ANTLRParser::_setmatch(
   SetWordType *tset
)
{
	if ( !set_el(LA(1), tset) ) {
      fail_syntax(tset);
	}
}

//
// Exception-throwing routines
//
////////////////////////////////////////////////////////////////
void 
ANTLRParser::fail_syntax(
   ANTLRTokenType tt
)
{
   // Make a token set consisting of just the desired token
   memset(tempTokenSet, 0, sizeof(tempTokenSet[0]) * tokenSetSize);
   tempTokenSet[DIVWORD(tt)] = bitmask[MODWORD(tt)];
   fail_syntax(tempTokenSet);
}

////////////////////////////////////////////////////////////////
void 
ANTLRParser::fail_syntax(
   SetWordType* set
)
{
   AParserEXSyntax ex(set);
   throw(ex);
}

////////////////////////////////////////////////////////////////
void 
ANTLRParser::fail_predicate(
   const char* msg
)
{
   AParserEXPredicate ex(msg);
   throw(ex);
}


//
// Exception handling routines
//

////////////////////////////////////////////////////////////////
void 
ANTLRParser::consumeUntil(
   SetWordType *st
)
{
	while ( !set_el(LA(1), st) && LA(1) != eofToken ) 
   { 
      consume(); 
   }
}

////////////////////////////////////////////////////////////////
void 
ANTLRParser::consumeUntilToken(
   int t
)
{
	while ( LA(1)!=t && LA(1) != eofToken ) 
   { 
      consume(); 
   }
}

////////////////////////////////////////////////////////////////
// is b an element of set p?
int 
ANTLRParser::set_el(
   ANTLRTokenType b, 
   SetWordType *p
) const
{
	return( p[DIVWORD(b)] & bitmask[MODWORD(b)] );
}


////////////////////////////////////////////////////////////////
// virtual 
void 
ANTLRParser::panic(
   char *msg
)
{
	fprintf(stdout, "ANTLR panic: %s\n", msg);
	assert(0);
}

////////////////////////////////////////////////////////////////
//virtual 
void 
ANTLRParser::trace_output(const char* msg)
{
   fprintf(stdout, "%s", msg);
	fflush(stdout);
}


