// AParserTrace.cpp: Implementation of tracer object

#include "AParser.h"

ANTLRParser::Tracer::Tracer(ANTLRParser* parser_, const char* msg_) :
   parser(parser_),
   msg(msg_)
{
   if ( !parser->isTracing() ) {
      return;
   }
   tabs(); 
   parser->trace_output("Enter ");
   OutputStuff(msg);
   level++; 
}

ANTLRParser::Tracer::~Tracer() 
{
   if ( !parser->isTracing() ) {
      return;
   }
   level--;
   tabs(); 
   parser->trace_output("Exit ");
   OutputStuff(msg);
}

void
ANTLRParser::Tracer::OutputStuff(
   const char* msg
) const
{
   parser->trace_output(msg); 
   parser->trace_output(": ");
   if (parser->guessing)
   {
      parser->trace_output("[guessing] ");
   }
   for (int i = 1; i <= 6; i++)
   {
      for (const char *p = parser->LT(i)->getText(); *p != 0; p++)
      {
         if (*p == '\n')
         {
            parser->trace_output("\\n");
         } else {
            char c[2];
            c[0] = *p;
            c[1] = 0;
            parser->trace_output(c);
         }
      }
      parser->trace_output(" ");
   }
   parser->trace_output("\n");
}

void 
ANTLRParser::Tracer::tabs() const 
{ 
   for (int i = 0; i < level; i++) 
   { 
      parser->trace_output("  "); 
   } 
}
