/* ANTLRToken.h
 *
 * SOFTWARE RIGHTS
 *
 * We reserve no LEGAL rights to the Purdue Compiler Construction Tool
 * Set (PCCTS) -- PCCTS is in the public domain.  An individual or
 * company may do whatever they wish with source code distributed with
 * PCCTS or the code generated by PCCTS, including the incorporation of
 * PCCTS, or its output, into commerical software.
 * 
 * We encourage users to develop software with PCCTS.  However, we do ask
 * that credit is given to us for developing PCCTS.  By "credit",
 * we mean that if you incorporate our source code into one of your
 * programs (commercial product, research project, or otherwise) that you
 * acknowledge this fact somewhere in the documentation, research report,
 * etc...  If you like PCCTS and have developed a nice tool with the
 * output, please mention that you developed it using PCCTS.  In
 * addition, we ask that this header remain intact in our source code.
 * As long as these guidelines are kept, we expect to continue enhancing
 * this system and expect to make other tools available as they are
 * completed.
 *
 * Terence Parr
 * Parr Research Corporation
 * with Purdue University and AHPCRC, University of Minnesota
 * 1989-1995
 *
 * ANTLR 1.33J
 * Mercilessly hacked by John Lilley 1996-97
 */

#ifndef INCL_ATOKEN_H
#define INCL_ATOKEN_H

#include "tokens.h"
#include "JLRefPtr.h"
#include <string.h>

/* must define what a char looks like; can make this a class too */
typedef char ANTLRChar;

class ANTLRAbstractToken : public RefCount {
public:

   ANTLRAbstractToken() {}
   virtual ~ANTLRAbstractToken() {}

   // Token type interface
   ANTLRTokenType getType() const { return (ANTLRTokenType)tokenType; }
   void setType(ANTLRTokenType tokenType_) { tokenType = (short)tokenType_; }

   // Line interface must be filled in by subclass
   virtual int getLine() const = 0;
   virtual void setLine(int line) = 0;

   // Text interface must be filled in by subclass
   virtual const ANTLRChar *getText() const = 0;
   virtual void setText(const ANTLRChar *) = 0;

   // Create a token with refcount==0
   // This supports the lexer creation of tokens.
   virtual ANTLRAbstractToken *makeToken(
      ANTLRTokenType tt,
      ANTLRChar *text,
      int line
   ) = 0;

   // Does not compare line or column info, just text and type
   bool operator==(const ANTLRAbstractToken& rhs) const
   {
      return
         getType() == rhs.getType() &&
         strcmp(getText(), rhs.getText()) == 0;
   }

protected:
   short tokenType;
};

typedef ANTLRAbstractToken *_ANTLRTokenPtr;

#endif
