/* DLexer.c
 *
 * SOFTWARE RIGHTS
 *
 * We reserve no LEGAL rights to the Purdue Compiler Construction Tool
 * Set (PCCTS) -- PCCTS is in the public domain.  An individual or
 * company may do whatever they wish with source code distributed with
 * PCCTS or the code generated by PCCTS, including the incorporation of
 * PCCTS, or its output, into commerical software.
 * 
 * We encourage users to develop software with PCCTS.  However, we do ask
 * that credit is given to us for developing PCCTS.  By "credit",
 * we mean that if you incorporate our source code into one of your
 * programs (commercial product, research project, or otherwise) that you
 * acknowledge this fact somewhere in the documentation, research report,
 * etc...  If you like PCCTS and have developed a nice tool with the
 * output, please mention that you developed it using PCCTS.  In
 * addition, we ask that this header remain intact in our source code.
 * As long as these guidelines are kept, we expect to continue enhancing
 * this system and expect to make other tools available as they are
 * completed.
 *
 * ANTLR 1.33
 * Terence Parr
 * Parr Research Corporation
 * with Purdue University and AHPCRC, University of Minnesota
 * 1989-1995
 */

// This must be a macro because ZZSHIFT is in the generated lexer code
#define ZZGETC { ch = buffer.nextChar(); cl = ZZSHIFT(ch);}

// This must be a macro to use the generated dfa members
#define ZZNEWSTATE {newstate = dfa[state][cl];}

void DLGLexer::
mode( int m )
{
	/* points to base of dfa table */
	if (m<MAX_MODE){
		automaton = m;
		/* have to redo class since using different compression */
		cl = ZZSHIFT(ch);
	}else{
		sprintf((char *)ebuf,"Invalid automaton mode = %d ",m);
		errstd(ebuf);
	}
}

ANTLRTokenType DLGLexer::
nextTokenType(void)
{
	register int state, newstate;
	ANTLRTokenType tk;

skip:
	nextpos = _lextext;
	_begcol = _endcol+1;
more:
	_begexpr = nextpos;

	/* interactive-only version of automaton */
	/* if there is something in ch, process it */
	state = newstate = dfa_base[automaton];
	if (charfull){
      ++_endcol; 
		CopyCharToLexText();
		ZZNEWSTATE;
	}
	while (alternatives[newstate]){
		state = newstate;
		ZZGETC;
      ++_endcol; 
		CopyCharToLexText();
		ZZNEWSTATE;
	}
	/* figure out if last character really part of token */
	if ((state != dfa_base[automaton]) && (newstate == DfaStates))
   {
		charfull = 1;
		--nextpos;
	} else {
		charfull = 0;
		state = newstate;
	}

   // Terminate the buffer
	*(nextpos) = '\0';

	/* Able to transition out of start state to some non err state?*/
	if ( state == dfa_base[automaton] ){
		/* make sure doesn't get stuck */
		advance();
	}

	_endcol -= charfull;
	_endexpr = nextpos -1;
	add_erase = 0;

   // Execute the next action and get the resulting token
	tk = (this->*actions[accepts[state]])();

	switch (add_erase) {
		case 1: goto skip;
		case 2: goto more;
	}

	return tk;
}

void 
DLGLexer::advance()
{
   assert(input != NULL);
	ZZGETC; 
   charfull = 1; 
   ++_endcol; 
}
