/* DLGLexerBase.c
 *
 * SOFTWARE RIGHTS
 *
 * We reserve no LEGAL rights to the Purdue Compiler Construction Tool
 * Set (PCCTS) -- PCCTS is in the public domain.  An individual or
 * company may do whatever they wish with source code distributed with
 * PCCTS or the code generated by PCCTS, including the incorporation of
 * PCCTS, or its output, into commerical software.
 * 
 * We encourage users to develop software with PCCTS.  However, we do ask
 * that credit is given to us for developing PCCTS.  By "credit",
 * we mean that if you incorporate our source code into one of your
 * programs (commercial product, research project, or otherwise) that you
 * acknowledge this fact somewhere in the documentation, research report,
 * etc...  If you like PCCTS and have developed a nice tool with the
 * output, please mention that you developed it using PCCTS.  In
 * addition, we ask that this header remain intact in our source code.
 * As long as these guidelines are kept, we expect to continue enhancing
 * this system and expect to make other tools available as they are
 * completed.
 *
 * ANTLR 1.33
 * Terence Parr
 * Parr Research Corporation
 * with Purdue University and AHPCRC, University of Minnesota
 * 1989-1995
 */
#include <stdio.h>
#include <stdlib.h>

#include "tokens.h"
#include "assert.h"

//#define ANTLR_SUPPORT_CODE
//#include "config.h"

#include "DLexerBase.h"

DLGLexerBase::DLGLexerBase(
   DLGInputStream *in
) : 
   buffer(in),
   _bufsize(DefaultBufSize),
	_begexpr(0),
   _endexpr(0),
	ch(0),
	cl(0),
	add_erase(0),
	input(in),
	_begcol(0),
	_endcol(0),
	_line(1),
	charfull(0),
	automaton(0),
	token_to_fill(0)
{
	_lextext = new DLGChar[_bufsize];
   nextpos = _lextext;
   *nextpos = (DLGChar)0;
	lastpos = &_lextext[_bufsize-1];
}

void DLGLexerBase::
setInputStream( DLGInputStream *in )
{
	input = in;
	_line = 1;
	charfull = 0;
}

/* saves dlg state, but not what feeds dlg (such as file position) */
void DLGLexerBase::
saveState(DLGState *state)
{
	state->input = input;
	state->auto_num = automaton;
	state->add_erase = add_erase;
	state->lookc = ch;
	state->char_full = charfull;
	state->begcol = _begcol;
	state->endcol = _endcol;
	state->line = _line;
	state->lextext = _lextext;
	state->begexpr = _begexpr;
	state->endexpr = _endexpr;
	state->bufsize = _bufsize;
	state->nextpos = nextpos;
	state->lastpos = lastpos;
	state->class_num = cl;
}

void DLGLexerBase::
restoreState(DLGState *state)
{
	input = state->input;
	automaton = state->auto_num;
	add_erase = state->add_erase;
	ch = state->lookc;
	charfull = state->char_full;
	_begcol = state->begcol;
	_endcol = state->endcol;
	_line = state->line;
	_lextext = state->lextext;
	_begexpr = state->begexpr;
	_endexpr = state->endexpr;
	_bufsize = state->bufsize;
	nextpos = state->nextpos;
	lastpos = state->lastpos;
	cl = state->class_num;
}

/* erase what is currently in the buffer, and get a new reg. expr */
void DLGLexerBase::
skip()
{
	add_erase = 1;
}

/* don't erase what is in the lextext buffer, add on to it */
void DLGLexerBase::
more()
{
	add_erase = 2;
}

/* substitute c for the reg. expr last matched and is in the buffer */
void DLGLexerBase::
replchar(DLGChar c)
{
   DLGChar s[2];
   s[0] = c;
   s[1] = 0;
   replstr(s);
}

/* replace the string s for the reg. expr last matched and in the buffer */
void DLGLexerBase::
replstr(register DLGChar *s)
{
	nextpos = _begexpr;

   if (s != 0)
   {
      // Copy the string into the buffer
      for (; *s != 0; s++)
      {
         if (nextpos + 1 >= lastpos)
         {
            extend_lextext();
         }
         *nextpos++ = *s;
      }
   }

   // null-terminate string
	*(nextpos) = '\0';

   // Record the end of the current text
	_endexpr = nextpos - 1;
}

//virtual 
void 
DLGLexerBase::errout(char*s)
{
   fprintf(stderr, "%s\n", s);
}

void 
DLGLexerBase::errstd(char *s)
{
   char *buf = new char[strlen(s) + strlen(_lextext) + 100];
   sprintf(
      buf,
      "%s near line %d (text was '%s')",
      ((s == NULL) ? "Lexical error" : s),
      _line, _lextext
   );
   errout(buf);
   delete [] buf;
}

int DLGLexerBase::
err_in()
{
	errout("No input stream, function, or string");
	/* return eof to get out gracefully */
	return EOF;
}

ANTLRTokenType 
DLGLexerBase::erraction()
{
	errstd("invalid token");
	advance();
	skip();
	return (ANTLRTokenType) 0;	// bogus, but satisfies compiler
}

ANTLRTokenPtr 
DLGLexerBase::getToken()
{
	assert(token_to_fill !=NULL );

   // Do not make this call a parameter of makeToken because _lextext may
   // be pushed on the arg stack before extend_lextext() is called.
   ANTLRTokenType tt = nextTokenType();

	return token_to_fill->makeToken(
   	tt,
      _lextext,
      _line
   );
}

//virtual
void 
DLGLexerBase::panic(char *msg)
{
	fprintf(stderr, "DLG panic: %s\n", msg);
	exit(EXIT_FAILURE);
}

void 
DLGLexerBase::extend_lextext()
{
   // Double the buffer size
   DLGChar* newText = new DLGChar[_bufsize * 2];

   // Copy old buffer to new
   for (int i = 0; i < _bufsize; i++)
   {
      newText[i] = _lextext[i];
   }

   // Point data members into the new buffer
	nextpos = newText + (nextpos - _lextext);
	_begexpr = newText + (_begexpr - _lextext);
	_endexpr = newText + (_endexpr - _lextext);

   // Delete the old buffer
   delete [] _lextext;

   // Set the new buffer
   _lextext = newText;
   _bufsize = _bufsize * 2;
	lastpos = &_lextext[_bufsize-1];
   *lastpos = (DLGChar)0;
}
