#header <<#include "MyZZPred.h">>

<<
/* user must define ANTLRToken */
class ANTLRToken : public ANTLRAbstractToken {
protected:
    ANTLRTokenType _type;	// what's the token type of the token object
    int _line;			// track line info for errors

	/* For our simple purposes, a token and a string is enough for
	 * our attribute
	 */
	ANTLRChar _text[30];

public:
	ANTLRToken(ANTLRTokenType t, ANTLRChar *s)
        { setType(t); _line = 0; setText(s); }

	/* Your derived class MUST have a blank constructor. */
    ANTLRToken()
		{ setType((ANTLRTokenType)0); _line = 0; setText(""); }

	// how to access the token type and line number stuff
   virtual int getLine() const { return _line; }
   void setLine(int line) { _line = line; }

	const ANTLRChar *getText() const { return _text; }
	void setText(const ANTLRChar *s) { strncpy(_text, s, 30); }

	/* WARNING WARNING WARNING: you must return a stream of distinct tokens */
	/* This function will disappear when I can use templates */
	virtual ANTLRAbstractToken *makeToken(ANTLRTokenType tt,
										  ANTLRChar *txt,
										  int line)
		{
			ANTLRAbstractToken *t = new ANTLRToken(tt,txt);
			t->setLine(line);
			return t;
		}
};

/* "DLGLexer" must match what you use on DLG command line (-cl);
 * "DLGLexer" is the default.
 */
#include "DLGLexer.h"		/* include definition of DLGLexer.
							 * This cannot be generated automatically because
							 * ANTLR has no idea what you will call this file
							 * with the DLG command-line options.
							 */

int main()
{
	DLGFileInput in(stdin);	/* create input stream for DLG to get chars from */
	DLGLexer scan(&in);		/* create scanner reading from stdin */
	ANTLRTokenBuffer pipe(&scan, 2);/* make buffered pipe between lexer&parser */
	ANTLRTokenPtr aToken=new ANTLRToken; // create a token to fill in for DLG
	scan.setToken(mytoken(aToken));
	Expr parser(&pipe);		/* create parser of type Expr hooked to scanner */
	parser.init();			/* init the parser; prime lookahead etc... */

	parser.r();				/* start parsing at rule 'r' of that parser */
	return 0;
}
>>


#token Eof "@"
#token "[\r \n \t \ ]"   << skip(); >>
#token A "[Aa]"
#token B "B"
#token C "C"
#token D "D"

class Expr {
<< 
   int predicate() { return 1; }
>>

/*
r : 
   A B C Eof
   ;
   exception << fprintf(stderr, "Caught error in rule\n"); >>
*/

r : (A)? B;

/*
r : ((ac)? | a )+ Eof ;
r2 : ( ac )* a;
r3 : ( (a)? | b )+;
r4 : ( a )*;
r5 : ( a | b )+;
r6 : ( a | b )*;
r7 : ( (a)? | b )*;
r8 : { a };
r9 : { (a)? | b };
r10 : <<predicate()>>? a;
r11 : ( (a)? )+;
r12 : ( (a)? )*;
r13 : ( ac )+ a;
r14 : (a)? | (b)? | c ;
r15 : (ac)? ((c)?) ;
r16 : (ac c)? ;
r17 : ( (ac c)? )+;
a : A;
ac : A C;
little_a : <<*LT(1)->getText()=='a'>>? A;
b : B;
c : C;
d : D;
*/

}
